import numpy as np
import os
import os.path as osp
import torch
import torch.nn as nn
import torch.nn.functional as F
import json
from typing import Callable
from .dataprep import SampleMixUp
from .utils import compute_entropy_of_mixing
from .model import CellTypeCLF, DANN
import copy
from torch.utils.tensorboard import SummaryWriter


class Trainer(object):
    '''
    Trains a PyTorch model.

    Attributes
    ----------
    model : nn.Module
        model with required `.forward(...)` method.
    criterion : Callable
        loss criterion to optimize.
    optimizer : torch.optim.Optimizer
        optimizer for the model parameters.
    dataloaders : dict
        keyed by ['train', 'val'] with values corresponding
        to `torch.utils.data.DataLoader` for training
        and validation sets.
    out_path : str
        output path for best model.
    n_epochs : int
        number of epochs for training.
    min_epochs : int
        minimum number of epochs before saving weights.
    patience : int
        maximum number of epochs to wait before early stopping.
        if `None`, infinite patience is used (up to `n_epochs`).
    waiting_time : int
        number of epochs since the last best val loss.
    reg_criterion : Callable
        criterion to penalize layer weights.
    use_gpu : bool
        use CUDA acceleration.
    verbose : bool
        write all batch losses to stdout.
    save_freq : int
        Number of epochs between model checkpoints. Default = 10.
    scheduler : learning rate scheduler.
    '''

    def __init__(
        self,
        model: nn.Module,
        criterion: Callable,
        optimizer: torch.optim.Optimizer,
        dataloaders: dict,
        out_path: str,
        batch_transformers: dict = {},
        n_epochs: int = 50,
        min_epochs: int = 0,
        patience: int = None,
        exp_name: str = '',
        reg_criterion: Callable = None,
        use_gpu: bool = torch.cuda.is_available(),
        verbose: bool = False,
        save_freq: int = 10,
        scheduler: torch.optim.lr_scheduler = None,
        tb_writer: str = None,
    ) -> None:
        '''
        Trains a PyTorch `nn.Module` object provided in `model`
        on training and testing sets provided in `dataloaders`
        using `criterion` and `optimizer`.

        Saves model weight snapshots every `save_freq` epochs and saves the
        weights with the best testing loss at the end of training.

        Parameters
        ----------
        model : nn.Module
            model with required `.forward(...)` method.
        criterion : Callable
            loss criterion to optimize.
        optimizer : torch.optim.Optimizer
            optimizer for the model parameters.
        dataloaders : dict
            keyed by ['train', 'val'] with values corresponding
            to `torch.utils.data.DataLoader` for training
            and validation sets.
        out_path : str
            output path for best model.
        batch_transformers : dict
            apply transforms to minibatch inputs and targets.
            keys are ['train', 'val'], values are Callable.
        n_epochs : int
            number of epochs for training.
        min_epochs : int
            minimum number of epochs before saving weights.
        patience : int
            maximum number of epochs to wait before early stopping.
            if `None`, infinite patience is used (up to `n_epochs`).
        reg_criterion : callable
            criterion to penalize layer weights.
        use_gpu : bool
            use CUDA acceleration.
        verbose : bool
            write all batch losses to stdout.
        save_freq : int
            Number of epochs between model checkpoints. Default = 10.
        scheduler : torch.optim.lr_scheduler
            learning rate schedule.
            
        Returns
        -------
        None.
        '''
        self.model = model
        self.optimizer = optimizer
        self.criterion = criterion
        self.n_epochs = n_epochs
        self.min_epochs = min_epochs
        self.patience = patience if patience is not None else n_epochs
        self.waiting_time = 0
        self.dataloaders = dataloaders
        self.batch_transformers = batch_transformers
        self.out_path = out_path
        self.use_gpu = use_gpu
        self.verbose = verbose
        self.save_freq = save_freq
        self.best_acc = 0.
        self.best_loss = 1.0e10
        self.scheduler = scheduler
        self.reg_criterion = reg_criterion
        if tb_writer is not None:
            self.tb_writer = SummaryWriter(log_dir=tb_writer)
            os.makedirs(tb_writer, exist_ok=True)
        else:
            self.tb_writer = None

        if not os.path.exists(self.out_path):
            os.mkdir(self.out_path)
        # initialize log

        self.log_path = os.path.join(
            self.out_path, '_'.join([exp_name, 'log.csv']))
        with open(self.log_path, 'w') as f:
            header = 'Epoch,Running_Loss,Mode\n'
            f.write(header)

        self.parameters = {
            'out_path': out_path,
            'exp_name': exp_name,
            'n_epochs': n_epochs,
            'use_cuda': self.use_gpu,
            'train_batch_size': self.dataloaders['train'].batch_size,
            'val_batch_size': self.dataloaders['val'].batch_size,
            'train_batch_sampler': str(
                type(self.dataloaders['train'].sampler)),
            'val_batch_sampler': str(
                type(self.dataloaders['val'].sampler)),
            'optimizer_type': str(type(self.optimizer)),
            'learning_rate': self.optimizer.param_groups[0]['lr'],
            'model_hidden': self.model.n_hidden,
            'model_ngenes': self.model.n_genes,
            'model_ncelltypes': self.model.n_cell_types,
        }

        # write the log file header
        with open(self.log_path, 'w') as f:
            header = 'Epoch,Iter,Running_Loss,Mode\n'
            f.write(header)

    def train_epoch(self):
        '''Perform training across one full iteration through
        the data.
        '''
        self.model.train(True)
        i = 0
        running_loss = 0.0
        running_corrects = 0.
        running_total = 0.

        btrans = self.batch_transformers.get('train', None)
        for data in self.dataloaders['train']:
            # if a batch transformer is present,
            # transform the data before use
            if btrans is not None:
                data = btrans(data)

            inputs = data['input']
            labels = data['output']  # one-hot

            if self.use_gpu:
                inputs = inputs.cuda()
                labels = labels.cuda()
            else:
                pass
            inputs.requires_grad_()
            labels.requires_grad = False

            # zero gradients
            self.optimizer.zero_grad()

            # forward pass
            outputs = self.model(inputs)
            # predictions are the output nodes with
            # the highest values
            _, predictions = torch.max(outputs, 1)

            # remake an integer version of the labels for quick checking
            int_labels = torch.argmax(labels, 1)

            correct = torch.sum(predictions.detach() == int_labels.detach())

            # compute loss
            if self.reg_criterion is not None:
                reg_loss = self.reg_criterion(self.model)
                loss = self.criterion(outputs, labels) + reg_loss
            else:
                loss = self.criterion(outputs, labels)

            if self.verbose:
                print('batch loss: ', loss.item())
            if np.isnan(loss.data.cpu().numpy()):
                raise RuntimeError('NaN loss encountered in training')

            # compute gradients in a backward pass, update parameters
            loss.backward()
            self.optimizer.step()

            # statistics update
            running_loss += loss.item() / inputs.size(0)
            running_corrects += float(correct.item())
            running_total += float(labels.size(0))

            if i % 100 == 0 and self.verbose:
                print('Iter : ', i)
                print('running_loss : ', running_loss / (i + 1))
                print('running_acc  : ', running_corrects/running_total)
                print('corrects: %f | total: %f' %
                      (running_corrects, running_total))
                # append to log
                with open(self.log_path, 'a') as f:
                    f.write(str(self.epoch) + ',' + str(i) + ',' +
                            str(running_loss / (i + 1)) + ',train\n')
            i += 1

        epoch_loss = running_loss / len(self.dataloaders['train'])
        epoch_acc = running_corrects / running_total

        # append to log
        with open(self.log_path, 'a') as f:
            f.write(str(self.epoch) + ',' + str(i) + ',' +
                    str(running_loss / (i + 1)) + ',train_epoch\n')
        
        if self.tb_writer is not None:
            self.tb_writer.add_scalar('Loss/train', epoch_loss, self.epoch)
            self.tb_writer.add_scalar('Acc/train', epoch_acc, self.epoch)
            for i, p in enumerate(self.model.parameters()):
                self.tb_writer.add_histogram(
                    f'Grad/param{i:04}', p.grad, self.epoch,
                )

            self.tb_writer.add_scalar(
                'lr/lr',
                self.optimizer.state_dict()['param_groups'][0]['lr'],
                self.epoch,
            )
        
        if self.verbose:
            print('{} Loss : {:.4f}'.format('train', epoch_loss))
            print('{} Acc : {:.4f}'.format('train', epoch_acc))
            print('TRAIN EPOCH corrects: %f | total: %f' %
                  (running_corrects, running_total))

    @torch.no_grad()
    def val_epoch(self):
        '''Perform a pass through the validation data.
        Do not record gradients to speed things up.
        '''
        self.model.train(False)
        i = 0
        running_loss = 0.0
        running_corrects = 0
        running_total = 0

        btrans = self.batch_transformers.get('val', None)
        for data in self.dataloaders['val']:
            # if a batch transformer is present,
            # transform the data before use
            if btrans is not None:
                data = btrans(data)

            inputs = data['input']
            labels = data['output']  # one-hot
            if self.use_gpu:
                inputs = inputs.cuda()
                labels = labels.cuda()
            else:
                pass

            # zero gradients
            self.optimizer.zero_grad()
            # forward pass
            outputs = self.model(inputs)
            _, predictions = torch.max(outputs, 1)

            # remake an integer version of the labels for quick checking
            int_labels = torch.argmax(labels, 1)
            correct = torch.sum(predictions.detach() == int_labels.detach())
            if self.verbose > 1:
                print('PRED\n', predictions[:10, ...])
                print('LABEL\n', int_labels[:10, ...])
                print('CORRECT: ', correct)

            if self.reg_criterion is not None:
                reg_loss = self.reg_criterion(self.model)
                loss = self.criterion(outputs, labels) + reg_loss
            else:
                loss = self.criterion(outputs, labels)

            # statistics update
            running_loss += loss.item() / inputs.size(0)
            running_corrects += int(correct.item())
            running_total += int(labels.size(0))

            if i % 1 == 10 and self.verbose > 1:
                print('Iter : ', i)
                print('running_loss : ', running_loss / (i + 1))
                print('running_acc  : ', running_corrects/running_total)
                print('corrects: %f | total: %f' %
                      (running_corrects, running_total))
                # append to log
                with open(self.log_path, 'a') as f:
                    f.write(str(self.epoch) + ',' + str(i) + ',' +
                            str(running_loss / (i + 1)) + ',val\n')
            i += 1

        epoch_loss = running_loss / len(self.dataloaders['val'])
        epoch_acc = running_corrects / running_total
        # append to log
        with open(self.log_path, 'a') as f:
            f.write(
                str(self.epoch)
                + ','
                + str(i)
                + ','
                + str(running_loss / (i + 1))
                + ',val_epoch\n')

        # add one epoch to the waiting time for best loss
        # if we had a new best loss, the counter is reset below
        self.waiting_time += 1
        if (epoch_loss < self.best_loss) and (self.epoch >= self.min_epochs):
            self.best_loss = epoch_loss
            self.best_model_wts = self.model.state_dict()
            self.waiting_time = 0
            torch.save(
                self.model.state_dict(),
                os.path.join(self.out_path,
                             ('model_weights_'
                              + str(self.epoch).zfill(3)
                              + '.pkl')
                             )
            )
            print('Saving best model weights...')
            torch.save(
                self.model.state_dict(),
                os.path.join(self.out_path, '00_best_model_weights.pkl'),
            )
            print('Saved best weights.')
            
            if hasattr(self, 'dan_criterion'):
                print('Trainer has a `dan_criterion`.')
                if self.dan_criterion is not None:
                    print('Saving DAN weights...')
                    torch.save(
                        self.dan_criterion.dann.state_dict(),
                        os.path.join(
                            self.out_path,
                            '02_best_dan_weights.pkl',
                        )
                    )
            
            with open(self.log_path, 'a') as f:
                f.write(
                    str(self.epoch)
                    + ','
                    + str(i)
                    + ','
                    + str(running_loss / (i + 1))
                    + ',best_model_weights\n',
                )
                
            if self.tb_writer is not None:
                self.tb_writer.add_text(
                    'BestWeights',
                    f'Saved best weights at {self.epoch}, loss {epoch_loss}',
                    self.epoch,
                )
                self.tb_writer.flush()
            
        elif (self.epoch % self.save_freq == 0):
            torch.save(
                self.model.state_dict(),
                os.path.join(
                    self.out_path,
                    'model_weights_'
                    + str(self.epoch).zfill(3)
                    + '.pkl',
                ),
            )
        
        elif self.epoch == (self.n_epochs-1):
            torch.save(
                self.model.state_dict(),
                os.path.join(
                    self.out_path,
                    '01_final_model_weights.pkl'
                ),
            )
        if self.verbose:
            print(f'{self.waiting_time} epochs since last best weights.\n')
        
        if self.tb_writer is not None:
            self.tb_writer.add_scalar('Loss/val', epoch_loss, self.epoch)
            self.tb_writer.add_scalar('Acc/val', epoch_acc, self.epoch)
            self.tb_writer.flush()
        
        if self.verbose:
            print('{} Loss : {:.4f}'.format('val', epoch_loss))
            print('{} Acc : {:.4f}'.format('val', epoch_acc))
            print('VAL EPOCH corrects: %f | total: %f' %
                  (running_corrects, running_total))

    def train(self):
        for epoch in range(self.n_epochs):
            self.epoch = epoch
            msg = f'Epoch {epoch}/{self.n_epochs-1}'
            p_complete = epoch/self.n_epochs
            n_bars = int(np.floor(30*p_complete))
            msg += '|' + '-'*n_bars + '_'*(30-n_bars) + '|'
            # print a new line so the progress bar isn't overwritten
            # on the final stdout
            end_char = '\n' if epoch == (self.n_epochs-1) else '\r'
            print(msg, end=end_char)
            
            # training epoch
            self.train_epoch()
            # evaluate model
            self.val_epoch()
            
            # update learning rate
            # NOTE: change in `torch>=1.1.0`, `scheduler.step()`
            # is now called AFTER `optimizer.step()`
            if self.scheduler is not None:
                self.scheduler.step()
                
            if self.waiting_time > self.patience:
                # we have waited a sufficient number of epochs
                # to perform early stopping
                print('>'*5)
                print(f'Early stopping at epoch {self.epoch}')
                print('>'*5)
                break

        self.model.load_state_dict(self.best_model_wts)
        
        if self.tb_writer is not None:
            # close tensorboard writer
            self.tb_writer.flush()
            self.tb_writer.close()
        
        return self.model


class SemiSupervisedTrainer(Trainer):

    def __init__(
        self,
        unsup_criterion: Callable,
        unsup_dataloader: torch.utils.data.DataLoader,
        unsup_weight: Callable,
        dan_criterion: Callable=None,
        dan_weight: Callable=None,
        **kwargs,
    ) -> None:
        '''Train a PyTorch model using both a supervised and
        unsupervised loss as described for Interpolation
        Consistency Training.

        Parameters
        ----------
        unsup_criterion : Callable
            loss function for unlabeled samples.
            takes both the current `nn.Module` model and a `torch.FloatTensor`
            of unlabeled samples as input.
        unsup_dataloader : torch.utils.data.DataLoader
            data loader supplying unlabeled samples.
        unsup_weight : Callable
            takes an int epoch as input and returns a weight coefficient
            to scale the importance of the unsupervised loss.
        dan_criterion : Callable, optional
            domain adaptation loss. takes in a model, labeled batch, and
            unlabeled batch, and returns a `torch.Tensor` loss value.
        dan_weight : Callable, optional
            domain adaptation loss weight schedule.
            takes an int epoch as input and returns a weight coefficient.

        Returns
        -------
        None.
        '''
        super(SemiSupervisedTrainer, self).__init__(**kwargs)
        self.unsup_criterion = unsup_criterion
        self.unsup_dataloader = unsup_dataloader
        self.unsup_weight = unsup_weight
        self.dan_criterion = dan_criterion
        if self.dan_criterion is not None:
            print('Using a Domain Adaptation Loss.')
        self.dan_weight = dan_weight
        return

    def train_epoch(self,) -> None:
        '''
        Perform training using both a supervised and semi-supervised loss.

        Notes
        -----
        (1) Sample labeled examples, compute the standard supervised loss.
        (2) Sample unlabeled examples, compute unsupervised loss.
        (3) Perform backward pass and update parameters.
        '''
        self.model.train(True)
        i = 0
        running_loss = 0.
        running_sup_loss = 0. # supervised loss
        running_uns_loss = 0. # unsupervised loss
        running_dom_loss = 0. # domain adaptation loss
        running_corrects = 0.
        running_total = 0.

        btrans = self.batch_transformers.get('train', None)

        iter_unsup_dl = iter(self.unsup_dataloader)
        for data in self.dataloaders['train']:

            ####################################
            # (1) Prepare data and graph
            ####################################
            
            # get unlabeled batch
            unsup_data = next(iter_unsup_dl)
            
            if btrans is not None:
                data = btrans(data)
            
            if self.use_gpu:
                # push all the data to the CUDA device
                data['input'] = data['input'].cuda()
                data['output'] = data['output'].cuda()
                
                unsup_data['input'] = unsup_data['input'].cuda()
        
            # capture gradients on labeled and unlabeled inputs
            # do not store gradients on labels
            data['input'].requires_grad = True
            data['output'].requires_grad = False
            
            unsup_data['input'].requires_grad = True
            
            # zero gradients across the graph
            self.optimizer.zero_grad()

            ####################################
            # (2) Compute loss terms
            ####################################

            sup_loss, unsup_loss, sup_outputs = self.unsup_criterion(
                model=self.model,
                unlabeled_sample=unsup_data,
                labeled_sample=data,
            )
            
            # check supervised classification accuracy
            _, predictions = torch.max(sup_outputs, 1)
            int_labels = torch.argmax(data['output'], 1)

            correct = torch.sum(predictions.detach() == int_labels.detach())

            # compute regularization loss
            if self.reg_criterion is not None:
                reg_loss = self.reg_criterion(self.model)
            else:
                reg_loss = 0.
                
            # compute the domain adaptation loss if desired
            if self.dan_criterion is not None:
                dan_weight = self.dan_weight(self.epoch)
                # NOTE: pseudolabel confidence is only used if `use_conf_pseudolabels`
                # was passed to the initiatilization of `DANLoss`
                pseudolabel_confidence = self.unsup_criterion.running_confidence_scores[-1][0]
                dan_loss = self.dan_criterion(
                    labeled_sample=data,
                    unlabeled_sample=unsup_data,
                    weight=dan_weight,
                    pseudolabel_confidence=pseudolabel_confidence,
                )
            else:
                dan_loss = torch.zeros(1,).float()
                dan_loss = dan_loss.to(device=sup_loss.device)
                dan_weight = 0.

            ####################################
            # (3) Perform backward pass
            ####################################

            loss = (
                sup_loss 
                + reg_loss 
                + (self.unsup_weight(self.epoch) * unsup_loss)
                + dan_loss
            )

            if self.verbose > 1:
                print('sup.  loss:   ', sup_loss.item())
                print('usup. loss:   ', unsup_loss.item())
                print('usup. weight: ', self.unsup_weight(self.epoch))
                if self.dan_criterion is not None:
                    print('Dom.  loss:   ', dan_loss.item())
                    print('Dom.  weight: ', dan_weight)
                print('total loss: ', loss.item())
            if np.isnan(loss.data.cpu().numpy()):
                raise RuntimeError('NaN loss encountered in training')

            # compute gradients in a backward pass, update parameters
            loss.backward()
            self.optimizer.step()

            # statistics update
            labeled_n = data['input'].size(0)
            unlabel_n = unsup_data['input'].size(0)
            
            running_loss += loss.item()
            running_sup_loss += sup_loss.item()
            running_uns_loss += unsup_loss.item()
            running_dom_loss += dan_loss.item()
            running_corrects += float(correct.item())
            running_total += float(data['input'].size(0))

            if i % 100 == 0 and self.verbose:
                print('Iter : ', i)
                print('running_sup_loss : ', running_sup_loss / (i + 1))
                print('running_uns_loss : ', running_uns_loss / (i + 1))
                print('running_dom_loss : ', running_dom_loss / (i + 1))                
                print('running_loss : ', running_loss / (i + 1))
                print('running_acc  : ', running_corrects/running_total)
                print('corrects: %f | total: %f' %
                      (running_corrects, running_total))
                # append to log
                with open(self.log_path, 'a') as f:
                    f.write(str(self.epoch) + ',' + str(i) + ',' +
                            str(running_loss / (i + 1)) + ',train\n')
            i += 1

        epoch_sup_loss = running_sup_loss / len(self.dataloaders['train'])
        epoch_uns_loss = running_uns_loss / len(self.dataloaders['train'])
        epoch_dom_loss = running_dom_loss / len(self.dataloaders['train'])
        epoch_loss = running_loss / len(self.dataloaders['train'])
        epoch_acc = running_corrects / running_total
        
        if self.tb_writer is not None:
            self.tb_writer.add_scalar(
                'Loss/train', epoch_loss, self.epoch,
            )
            self.tb_writer.add_scalar(
                'Acc/train', epoch_acc, self.epoch,
            )
            self.tb_writer.add_scalar(
                'Loss/super', epoch_sup_loss, self.epoch,
            )
            self.tb_writer.add_scalar(
                'Loss/unsup', epoch_uns_loss, self.epoch,
            )
            self.tb_writer.add_scalar(
                'SSL/UnsWeight', self.unsup_weight(self.epoch), self.epoch,
            )
            if self.dan_criterion is not None:
                self.tb_writer.add_scalar(
                    'Loss/domain', epoch_dom_loss, self.epoch,
                )            
                self.tb_writer.add_scalar(
                    'SSL/DomWeight', self.dan_weight(self.epoch), self.epoch,
                )
                
                # add embedding
                dlabel = self.dan_criterion.dlabel.numpy()
                self.tb_writer.add_embedding(
                    self.dan_criterion.x_embed,
                    metadata=dlabel.tolist(),
                    global_step=self.epoch,
                    tag='Embed/DAN',
                )
                
                # compute the entropy of mixing
                dan_embedding = self.dan_criterion.x_embed.numpy()
                
                eom = compute_entropy_of_mixing(
                    X=dan_embedding,
                    y=dlabel[:,0],
                    n_neighbors=100,
                    n_iters=512,
                    n_jobs=-1,
                )
                self.tb_writer.add_scalar(
                    'SSL/entropy_of_mixing', np.mean(eom), self.epoch,
                )
                self.tb_writer.add_histogram(
                    'SSL/dist_entropy_of_mixing', eom, self.epoch,
                )
                self.tb_writer.add_scalar(
                    'SSL/domain_acc', self.dan_criterion.dan_acc, self.epoch,
                )
                
                for i, param in enumerate(self.dan_criterion.dann.domain_clf.parameters()):
                    self.tb_writer.add_histogram(
                        f'Grad/domain_clf_{i:04}', param.grad, self.epoch,
                    )
                self.tb_writer.add_scalar(
                    'SSL/dan_n_conf_pseudolabels', 
                    self.dan_criterion.n_conf_pseudolabels,
                    self.epoch,
                )
                self.tb_writer.add_scalar(
                    'SSL/dan_p_conf_pseudolabels',
                    self.dan_criterion.n_conf_pseudolabels/self.dan_criterion.n_total_unlabeled,
                    self.epoch,                    
                )

            self.tb_writer.flush()
                   
            for i, named_mod in enumerate(self.model.classif.named_modules()):
                module_name = named_mod[0]
                module = named_mod[1]
                for j, param in enumerate(module.parameters()):
                    self.tb_writer.add_histogram(
                        f'Grad/{module_name}/{j:04}', param.grad, self.epoch,
                    )
            
            # add the running confidence scores of unlabeled examples
            # if we're using MixMatch
            if hasattr(self.unsup_criterion, 'running_confidence_scores'):
                # get the number of confident pseudolabels
                # and the total number of pseudolabels per batch
                n_conf = torch.Tensor([
                    torch.sum(s[0]).item() for s in self.unsup_criterion.running_confidence_scores
                ])
                n_total = torch.Tensor([
                    s[0].size(0) for s in self.unsup_criterion.running_confidence_scores
                ])
                conf_dist = torch.cat(
                    [
                        s[1] for s in self.unsup_criterion.running_confidence_scores
                    ],
                    dim=0,
                )
                self.tb_writer.add_scalar(
                    'SSL/p_conf_pseudolabels',
                    torch.sum(n_conf)/torch.sum(n_total),
                    self.epoch,
                )
                self.tb_writer.add_scalar(
                    'SSL/avg_pseudolabel_conf',
                    torch.mean(conf_dist),
                    self.epoch,
                )
                self.tb_writer.add_histogram(
                    'SSL/dist_p_conf_pseudolabels',
                    n_conf/n_total,
                    self.epoch,
                )
                self.tb_writer.add_histogram(
                    'SSL/pseudolabel_conf',
                    conf_dist,
                    self.epoch,
                )

        # append to log
        with open(self.log_path, 'a') as f:
            f.write(str(self.epoch) + ',' + str(i) + ',' +
                    str(epoch_loss) + ',train_epoch\n')
            # write out the supervised and unsupervised components
            # of loss separately
            f.write(str(self.epoch) + ',' + str(i) + ',' +
                    str(epoch_sup_loss) + ',train_epoch_sup\n')
            f.write(str(self.epoch) + ',' + str(i) + ',' +
                    str(epoch_uns_loss) + ',train_epoch_uns\n')
            f.write(str(self.epoch) + ',' + str(i) + ',' +
                    str(self.unsup_weight(self.epoch)) + ',train_epoch_uns_weight\n')
        if self.verbose:
            print('{} Sup. Loss : {:.6f}'.format('train', epoch_sup_loss))
            print('{} Unsup. Loss : {:.6f}'.format('train', epoch_uns_loss))
            print('{} Unsup. Weight : {:.6f}'.format(
                'train', self.unsup_weight(self.epoch)))
            if self.dan_criterion is not None:
                print('{} Dom.  Loss : {:.6f}'.format('train', epoch_dom_loss))
                print(f'train Dom.  Weight : {self.dan_weight(self.epoch)}')
            print('{} Loss : {:.4f}'.format('train', epoch_loss))
            print('{} Acc : {:.4f}'.format('train', epoch_acc))
            print('TRAIN EPOCH corrects: %f | total: %f' %
                  (running_corrects, running_total))
        return


'''Loss functions'''


def get_class_weight(
    y: np.ndarray,
) -> np.ndarray:
    '''Generate relative class weights based on the representation
    of classes in a label vector `y`

    Parameters
    ----------
    y : np.ndarray
        [N,] vector of class labels.

    Returns
    -------
    class_weight : np.ndarray
        [Classes,] vector of loss weight coefficients.
        if classes are `str`, returns weights in lexographically
        sorted order.

    '''
    # find all unique class in y and their counts
    u_classes, class_counts = np.unique(y, return_counts=True)
    # compute class proportions
    class_prop = class_counts/len(y)
    # invert proportions to get class weights
    class_weight = 1./class_prop
    # normalize so that the minimum value is 1.
    class_weight = class_weight / class_weight.min()
    return class_weight


def cross_entropy(
    pred_: torch.FloatTensor,
    label: torch.FloatTensor,
    class_weight: torch.FloatTensor = None,
    sample_weight: torch.FloatTensor = None,
    reduction: str='mean',
) -> torch.FloatTensor:
    '''Compute cross entropy loss for prediction outputs
    and potentially non-binary targets.

    Parameters
    ----------
    pred_ : torch.FloatTensor
        [Batch, C] model outputs.
    label : torch.FloatTensor
        [Batch, C] labels. may not necessarily be one-hot,
        but must satisfy simplex criterion.
    class_weight : torch.FloatTensor
        [C,] relative weights for each of the output classes.
        useful for increasing attention to underrepresented
        classes.
    reduction : str
        reduction method across the batch.

    Returns
    -------
    loss : torch.FloatTensor
        mean cross-entropy loss across the batch indices.

    Notes
    -----
    Crossentropy is defined as:

    .. math::

        H(P, Q) = -\Sum_{k \in K} P(k) log(Q(k))

    where P, Q are discrete probability distributions defined
    with a common support K.
    
    References
    ----------
    See for class weight computation:
    https://pytorch.org/docs/stable/nn.html#crossentropyloss
    '''
    if pred_.size() != label.size():
        msg = f'pred size {pred_.size()} not compatible with label size {label.size()}\n'
        raise ValueError(msg)
        
    if reduction.lower() not in ('mean', 'sum', 'none'):
        raise ValueError(f'{reduction} is not a valid reduction method.')
    
    # Apply softmax transform to predictions and log transform
    pred_log_sm = torch.nn.functional.log_softmax(pred_, dim=1)
    # Compute cross-entropy with the label vector
    samplewise_loss = -1 * torch.sum(label * pred_log_sm, dim=1)
    
    if sample_weight is not None:
        # weight individual samples using sample_weight
        # we squeeze into a single column in-case it had an
        # empty singleton dimension
        samplewise_loss *= sample_weight.squeeze()
        
    if class_weight is not None:
        class_weight = class_weight.to(label.device)
        # weight the losses
        # copy the weights across the batch to allow for elementwise
        # multiplication with the samplewise losses
        class_weight = class_weight.repeat(samplewise_loss.size(0), 1)
        # compute an [N,] vector of weights for each samples' loss
        weight_vec, _ = torch.max(
            class_weight * label, 
            dim=1,
        )
        
        samplewise_loss = samplewise_loss * weight_vec
    if reduction == 'mean':
        loss = torch.mean(samplewise_loss)
    elif reduction == 'sum':
        loss = torch.sum(samplewise_loss)
    else:
        loss = samplewise_loss
    return loss


class InterpolationConsistencyLoss(object):

    def __init__(
        self,
        unsup_criterion: Callable,
        sup_criterion: Callable,
        decay_coef: float = 0.9997,
        mean_teacher: bool = True,
        augment: Callable = None,
        teacher_eval: bool = True,
        teacher_bn_running_stats: bool = None,
        **kwargs,
    ) -> None:
        '''Computes an Interpolation Consistency Loss
        given a trained model and an unlabeled minibatch.

        Parameters
        ----------
        unsup_criterion : Callable
            loss criterion for similarity between "mixed-up" 
            "fake labels" and predicted labels for "mixed-up"
            samples.
        sup_criterion : Callable
            loss for samples with a primarily labeled component.       
        decay_coef : float
            decay coefficient for mean teacher parameter
            updates.
        mean_teacher : bool
            use a mean teacher model for interpolation consistency
            loss estimation.
        augment : Callable
            augments a batch of samples.
        teacher_eval : bool
            place teacher in evaluation mode, deactivating stochastic
            model components.
        teacher_bn_running_stats : bool
            use running statistics for batch normalization mean and
            variance. 
            if False, uses minibatch statistics.
            if None, uses setting of the student model batch norm layer.

        Returns
        -------
        None.

        Notes
        -----
        Instantiates a `SampleMixUp` class and passes any
        `**kwargs` to this class.

        Uses a "mean teacher" method by keeping a running 
        average of parameter sets used in the `__call__` 
        method.

        `decay_coef` taken from the Mean Teacher paper experiments
        on ImageNet.
        https://arxiv.org/abs/1703.01780

        Formalism:

        .. math::

            icl(u) = criterion( f(Mixup(u_i, u_j)), 
                                Mixup(f(u_i), f(u_j)) )

        References
        ----------
        1. Interpolation consistency training for semi-supervised learning
        2019, arXiv:1903.03825v3, stat.ML
        Vikas Verma, Alex Lamb, Juho Kannala, Yoshua Bengio

        2. Mean teachers are better role models: \
            Weight-averaged consistency targets improve \
            semi-supervised deep learning results
        2017, arXiv:1703.01780, cs.NE
        Antti Tarvainen, Harri Valpola
        '''
        self.unsup_criterion = unsup_criterion
        self.sup_criterion = sup_criterion
        self.decay_coef = decay_coef
        self.mean_teacher = mean_teacher
        if self.mean_teacher:
            print('IC Loss is using a mean teacher.')
        self.augment = augment
        self.teacher_eval = teacher_eval
        self.teacher_bn_running_stats = teacher_bn_running_stats

        # instantiate a callable MixUp operation
        self.mixup_op = SampleMixUp(**kwargs)

        self.teacher = None
        self.step = 0
        return

    def _update_teacher(self, model: nn.Module,) -> None:
        '''Update the teacher model based on settings'''
        if self.mean_teacher:
            if self.teacher is None:
                # instantiate the teacher with a copy
                # of the model
                self.teacher = copy.deepcopy(model,)
            else:
                self._update_teacher_params(model,)
        else:
            self.teacher = copy.deepcopy(model,)
        
        if self.teacher_eval:
            self.teacher = self.teacher.eval()
            
        if self.teacher_bn_running_stats is not None:
            # enforce our preference for teacher model batch 
            # normalization statistics
            for m in self.teacher.modules():
                if isinstance(m, nn.BatchNorm1d):
                    m.track_running_stats = self.teacher_bn_running_stats
        
        # check that our parameters are preserved
        if self.teacher_bn_running_stats is not None:
            # enforce our preference for teacher model batch 
            # normalization statistics
            for m in self.teacher.modules():
                if isinstance(m, nn.BatchNorm1d):
                    assert m.track_running_stats == self.teacher_bn_running_stats
            
        return

    def _update_teacher_params(self, model: nn.Module,) -> None:
        '''Update parameters in the teacher model using an
        exponential averaging method.

        Notes
        -----
        Logic derived from the Mean Teacher implementation
        https://github.com/CuriousAI/mean-teacher/
        '''
        # Per the mean-teacher paper, we use the global average
        # of parameter values until the exponential average is more effective
        # For a `decay_coef ~= 0.997`, this hand-off happens at ~step 333.
        alpha = min(1 - 1 / (self.step+1), self.decay_coef)
        # Perform in-place operations on the teacher parameters to average
        # with the new model parameters
        # Here, we're computing a simple weighted average where alpha is
        # the weight on past parameters, and (1 - alpha) is the weight on
        # new parameters
        zipped_params = zip(self.teacher.parameters(), model.parameters())
        for teacher_param, model_param in zipped_params:
            (teacher_param
             .data.mul_(alpha)
             .add_(1 - alpha, model_param.data))
        return

    def __call__(
        self,
        model: nn.Module,
        unlabeled_sample: dict,        
        labeled_sample: dict,
    ) -> torch.FloatTensor:
        '''Takes a model and set of unlabeled samples as input
        and computes the Interpolation Consistency Loss.

        Parameters
        ----------
        model : nn.Module
            model with parameters accessible via the `.parameters()`
            method.
        unlabeled_sample : dict
            input - torch.FloatTensor
                [Batch, Features] minibatch of unlabeled samples.
            output - torch.LongTensor
                zeros.
        labeled_sample : dict
            input - torch.FloatTensor
                [Batch, Features] minibatch of labeled examples.
            output - torch.LongTensor 
                one-hot labels.

        Returns
        -------
        supervised_loss : torch.FloatTensor
            supervised loss computed using `sup_criterion` between
            model predictions on mixed observations and true labels.
        unsupervised_loss : torch.FloatTensor
            unsupervised loss computed using `criterion` and the 
            interpolation consistency method.
        supervised_outputs : torch.FloatTensor
            [Batch, Classes] model outputs for augmented labeled examples.
            

        Notes
        -----
        Algorithm description:

        (0) Update the mean teacher.
        (1) Compute "fake labels" for unlabeled samples by performing
        a forward pass through the "mean teacher" and using the output
        as a representative label for the sample.
        (2) Perform a MixUp operation on unlabeled samples and their
        corresponding fake labels.
        (3) Compute the loss criterion between the mixed-up fake labels
        and the predicted fake labels for the mixed up samples.
        '''
        ###############################
        # (0) Update the mean teacher
        ###############################

        self._update_teacher(model,)

        ###############################
        # (1) Compute Fake Labels
        ###############################
        
        with torch.no_grad():
            fake_y = F.softmax(
                self.teacher(unlabeled_sample['input']),
                dim=1,
            )

        ###############################
        # (2) Perform MixUp and Forward
        ###############################

        unlabeled_sample['output'] = fake_y

        mixed_sample = self.mixup_op(unlabeled_sample)
        # move sample to model device if necessary
        mixed_sample['input'] = mixed_sample['input'].to(
            device=next(model.parameters()).device,
        )
        mixed_output = F.softmax(
            model(mixed_sample['input']),
        )
        assert mixed_output.requires_grad
        
        # set outputs as attributes for later access
        self.mixed_output = mixed_output
        self.mixed_sample = mixed_sample
        self.unlabeled_sample = unlabeled_sample

        ###############################
        # (3) Compute unsupervised loss
        ###############################

        icl = self.unsup_criterion(
            mixed_output,
            fake_y,
        )
        
        ###############################
        # (4) Compute supervised loss
        ###############################
        
        if self.augment is not None:
            labeled_sample = self.augment(labeled_sample)
        # move sample to the model device if necessary
        labeled_sample['input'] = labeled_sample['input'].to(
            device=next(model.parameters()).device,
        )
        labeled_sample['input'].requires_grad = True
        
        sup_outputs = model(labeled_sample['input'])
        sup_loss = self.sup_criterion(
            sup_outputs,
            labeled_sample['output'],
        )
        
        self.step += 1
        return sup_loss, icl, sup_outputs
    
    
def sharpen_labels(
    q: torch.FloatTensor,
    T: float=0.5,
) -> torch.FloatTensor:
    '''Reduce the entropy of a categorical label using a 
    temperature adjustment
    
    Parameters
    ----------
    q : torch.FloatTensor
        [N, C] pseudolabel.
    T : float
        temperature parameter.
    
    Returns
    -------
    q_s : torch.FloatTensor
        [C,] sharpened pseudolabel.
        
    Notes
    -----
    .. math::
    
        S(q, T) = q_i^{1/T} / \sum_j^L q_j^{1/T}

    '''
    if T == 0.:
        # equivalent to argmax
        _, idx = torch.max(q, dim=1)
        oh = torch.nn.functional.one_hot(
            idx,
            num_classes=q.size(1),
        )
        return oh
    
    if T == 1.0:
        # no-op
        return q
    
    q = torch.pow(q, 1. / T)
    q /= torch.sum(q, dim=1,).reshape(-1, 1)
    return q

    
class MixMatchLoss(InterpolationConsistencyLoss):
    '''Compute the MixMatch Loss given a batch of labeled
    and unlabeled examples.
    
    Attributes
    ----------
    n_augmentations : int
        number of augmentated samples to average across when
        computing pseudolabels.
        default = 2 from MixMatch paper.         
    T : float
        temperature parameter.
    augment_pseudolabels : bool
        perform augmentations during pseudolabel generation.
    pseudolabel_min_confidence : float
        minimum confidence to compute a loss for a given pseudolabeled
        example. examples below this confidence threshold will be given
        `0` loss. see the `FixMatch` paper for discussion.
    teacher : nn.Module
        teacher model for pseudolabeling.
    running_confidence_scores : list
        [n_batches_to_store,] (torch.Tensor, torch.Tensor,) of unlabeled
        example (Confident_Bool, BestConfidenceScore) tuples.
    n_batches_to_store : int
        determines how many batches to keep in `running_confidence_scores`.
    '''
    
    def __init__(
        self,
        n_augmentations: int=2,
        T: float=0.5,
        augment_pseudolabels: bool=True,
        pseudolabel_min_confidence: float=0.0,
        **kwargs,
    ) -> None:
        '''Compute the MixMatch Loss given a batch of labeled
        and unlabeled examples.
        
        Parameters
        ----------
        n_augmentations : int
            number of augmentated samples to average across when
            computing pseudolabels.
            default = 2 from MixMatch paper.         
        T : float
            temperature parameter.
        augment_pseudolabels : bool
            perform augmentations during pseudolabel generation.
        pseudolabel_min_confidence : float
            minimum confidence to compute a loss for a given pseudolabeled
            example. examples below this confidence threshold will be given
            `0` loss. see the `FixMatch` paper for discussion.
        
        Returns
        -------
        None.
        
        References
        ----------
        MixMatch: A Holistic Approach to Semi-Supervised Learning
        http://papers.nips.cc/paper/8749-mixmatch-a-holistic-approach-to-semi-supervised-learning
        
        FixMatch: https://arxiv.org/abs/2001.07685
        '''
        # inherit from IC loss, forcing the SampleMixUp to keep 
        # the identity of the dominant observation in each mixed sample
        super(MixMatchLoss, self).__init__(
            **kwargs, 
            keep_dominant_obs=True,
        )
        if not callable(self.augment):
            msg = 'MixMatch requires a Callable for augment'
            raise TypeError(msg)
        self.n_augmentations = n_augmentations
        self.augment_pseudolabels = augment_pseudolabels
        self.T = T
        
        self.pseudolabel_min_confidence = pseudolabel_min_confidence
        # keep a running score of the last 50 batches worth of pseudolabel
        # confidence outcomes
        self.n_batches_to_store = 50
        self.running_confidence_scores = []
        return
    
    
    @torch.no_grad()
    def _generate_labels(
        self,
        unlabeled_sample: dict,
    ) -> torch.FloatTensor:
        '''Generate labels by applying a set of augmentations
        to each unlabeled example and keeping the mean.
        
        Parameters
        ----------
        unlabeled_batch : dict
            "input" - [Batch, Features] minibatch of unlabeled samples.        
        '''
        # let the teacher model take guesses at the label for augmented
        # versions of the unlabeled observations
        raw_guesses = []
        for i in range(self.n_augmentations):
            to_augment = {
                'input': unlabeled_sample['input'].clone(),
                'output' : torch.zeros(1),
            }
            if self.augment_pseudolabels:
                # augment the batch before pseudolabeling
                augmented_batch = self.augment(to_augment)
            else:
                augmented_batch = to_augment
            # convert model guess to probability distribution `q`
            # with softmax, prior to considering it a label
            guess = F.softmax(
                self.teacher(augmented_batch['input']),
                dim=1,
            )
            raw_guesses.append(guess)

        # compute pseudolabels as the mean across all label guesses
        pseudolabels = torch.mean(
            torch.stack(
                raw_guesses, 
                dim=0,
            ),
            dim=0,
        )
        
        # before sharpening labels, determine if the labels are 
        # sufficiently confidence to use
        highest_conf, likeliest_class = torch.max(
            pseudolabels,
            dim=1,
        )
        # confident is a bool that we will use to decide if we should
        # keep loss from a given example or zero it out
        confident = highest_conf >= self.pseudolabel_min_confidence
        # store confidence outcomes in a running list so we can monitor
        # which fraction of pseudolabels are being used
        if len(self.running_confidence_scores) > self.n_batches_to_store:
            # remove the oldest batch
            self.running_confidence_scores.pop(0)

        # store tuples of (torch.Tensor, torch.Tensor)
        # (confident_bool, highest_conf_score)
        self.running_confidence_scores.append(
            (
                confident.detach().cpu(),
                highest_conf.detach().cpu(),
            ),
        )
        
        if self.T is not None:
            # sharpen labels
            pseudolabels = sharpen_labels(
                q=pseudolabels,
                T=self.T,
            )
        # ensure pseudolabels aren't attached to the
        # computation graph
        pseudolabels = pseudolabels.detach()
        
        return pseudolabels, confident
    
    
    def __call__(
        self,
        model: nn.Module,
        unlabeled_sample: dict,
        labeled_sample: dict,
    ) -> (torch.FloatTensor, torch.FloatTensor):
        '''
        Parameters
        ----------
        model : nn.Module
            model with parameters accessible via the `.parameters()`
            method.        
        unlabeled_sample : dict
            input - torch.FloatTensor
                [Batch, Features] minibatch of unlabeled samples.
            output - torch.LongTensor
                zeros.
        labeled_sample : dict
            input - torch.FloatTensor
                [Batch, Features] minibatch of labeled examples.
            output - torch.LongTensor 
                one-hot labels.
            
        Returns
        -------
        supervised_loss : torch.FloatTensor
            supervised loss computed using `sup_criterion` between
            model predictions on mixed observations and true labels.
        unsupervised_loss : torch.FloatTensor
            unsupervised loss computed using `criterion` between
            model predictions on mixed unlabeled observations 
            and pseudolabels generated as the mean
            across `n_augmentations` augmentation runs.
        supervised_outputs : torch.FloatTensor
            [Batch, Classes] model outputs for augmented labeled examples.
        '''
        
        ########################################
        # (0) Update the mean teacher
        ########################################

        self._update_teacher(model,)
        
        ########################################
        # (1) Generate labels for unlabeled data
        ########################################
        
        pseudolabels, pseudolabel_confidence = self._generate_labels(
            unlabeled_sample = unlabeled_sample,
        )
        # make sure pseudolabels match real label dtype
        # so that they can be concatenated
        pseudolabels = pseudolabels.to(
            dtype=labeled_sample['output'].dtype
        )
        
        ########################################
        # (2) Augment the labeled data
        ########################################
        
        labeled_sample = self.augment(
            labeled_sample,
        )
        
        ########################################
        # (3) Perform MixUp across both batches
        ########################################
        n_unlabeled_original = unlabeled_sample['input'].size(0)
        unlabeled_sample['output'] = pseudolabels
        
        # separate samples into confident and unconfident sample dicts
        # we only allow samples with confident pseudolabels to 
        # participate in the MixUp operation
        conf_unlabeled_sample = {}
        ucnf_unlabeled_sample = {}
        
        for k in unlabeled_sample.keys():
            conf_unlabeled_sample[k] = unlabeled_sample[k][pseudolabel_confidence]
            ucnf_unlabeled_sample[k] = unlabeled_sample[k][~pseudolabel_confidence]
        
        # unlabeled samples come BEFORE labeled samples
        # in the concatenated sample
        # NOTE: we only allow confident unlabeled samples
        # into the concatenated sample used for MixUp
        cat_sample = {
            k: torch.cat(
                [
                    conf_unlabeled_sample[k],
                    labeled_sample[k],
                ],
                dim=0,
            ) for k in ['input', 'output']
        }
        
        # mixup the concatenated sample
        # NOTE: dominant observations are maintained
        # by passing `keep_dominant_obs=True` in
        # `self.__init__`
        mixed_samples = self.mixup_op(
            cat_sample,
        )
        
        ########################################
        # (4) Forward pass for mixed samples
        ########################################
        
        # split the mixed samples based on the dominant
        # observation
        n_unlabeled = conf_unlabeled_sample['input'].size(0)
        unlabeled_m_ = mixed_samples['input'][:n_unlabeled]
        unlabeled_y_ = mixed_samples['output'][:n_unlabeled]
        
        labeled_m_ = mixed_samples['input'][n_unlabeled:]
        labeled_y_ = mixed_samples['output'][n_unlabeled:]
        
        # append low confidence samples to unlabeled_m_ and unlabeled_y_
        # this ensures that batch norm is still able to update it's
        # statistics based on batches from the train AND target domain
        unlabeled_m_ = torch.cat([
            unlabeled_m_,
            ucnf_unlabeled_sample['input'],
        ])
        unlabeled_y_ = torch.cat([
            unlabeled_y_,
            ucnf_unlabeled_sample['output'],
        ])          
        
        # perform a forward pass on mixed samples
        # NOTE: Our unsupervised criterion operates on post-softmax
        # probability vectors, so we transform the output here
        unlabeled_z_ = F.softmax(
            model(unlabeled_m_),
            dim=1,
        )
        # NOTE: Our supervised criterion operates directly on
        # logits and performs a `logsoftmax()` internally
        labeled_z_ = model(labeled_m_)
        
        ########################################
        # (5) Compute losses
        ########################################
        
        # compare mixed pseudolabels to the model guess
        # on the mixed input
        # NOTE: this returns an **unreduced** loss of size 
        # [Batch,] or [Batch, Classes] depending on the loss function
        unsupervised_loss = self.unsup_criterion(
            unlabeled_z_,
            unlabeled_y_,
        )
        # sum loss across classes if not reduced in the loss
        if unsupervised_loss.dim() > 1:
            unsupervised_loss = torch.sum(unsupervised_loss, dim=1)

        # scale the loss to 0 for all observations without confident pseudolabels
        # this allows the loss to slowly ramp up as labels become more confident
        scale_vec = torch.zeros_like(unsupervised_loss).float().to(
            device=unsupervised_loss.device
        )
        scale_vec[:n_unlabeled] += 1.
        unsupervised_loss = unsupervised_loss * scale_vec
        unsupervised_loss = torch.mean(unsupervised_loss)
        
        # compute model guess on the mixed supervised input
        # to the mixed labels
        # NOTE: we didn't allow non-confident pseudolabels 
        # into the MixUp, so this shouldn't propogate any
        # poor quality pseudolabel information
        supervised_loss = self.sup_criterion(
            labeled_z_,
            labeled_y_,
        )
        
        self.step += 1
        
        return supervised_loss, unsupervised_loss, labeled_z_
    
    
class DANLoss(object):
    '''Compute a domain adaptation network (DAN) loss.'''
    
    def __init__(
        self,
        dan_criterion: Callable,
        model: CellTypeCLF,
        use_conf_pseudolabels: bool=False,
        scale_loss_pseudoconf: bool=False,
        n_domains: int=2,
        **kwargs,
    ) -> None:
        '''Compute a domain adaptation network loss.
        
        Parameters
        ----------
        dan_criterion : Callable
            domain classification criterion `Callable(output, target)`.
        model : scnym.model.CellTypeCLF
            `CellTypeCLF` model to use for embedding.
        use_conf_pseudolabels : bool
            only use unlabeled observations with confident pseudolabels
            for discrimination. expects `pseudolabel_confidence` to be
            passed in the `__call__()` if so.
        scale_loss_pseudoconf : bool
            scale the weight of the gradients passed to both models based
            on the proportion of confident pseudolabels.
        n_domains : int
            number of domains of origin to predict using the adversary.
            
        Returns
        -------
        None.
        
        Notes
        -----
        **kwargs are passed to `scnym.model.DANN`
        
        See Also
        --------
        scnym.model.DANN
        '''
        self.dan_criterion = dan_criterion
        
        # build the DANN
        self.dann = DANN(
            model=model,
            n_domains=n_domains,
            **kwargs,
        )
        self.dann.domain_clf = self.dann.domain_clf.to(
            device=next(iter(model.parameters())).device,
        )
        # instantiate with small tensor to simplify downstream size
        # checking logic
        self.x_embed = torch.zeros((1,1))
        
        self.use_conf_pseudolabels = use_conf_pseudolabels
        self.scale_loss_pseudoconf = scale_loss_pseudoconf
        return
    
    def __call__(
        self,
        labeled_sample: dict,
        unlabeled_sample: dict,
        weight: float,
        pseudolabel_confidence: torch.Tensor=None,
    ) -> torch.FloatTensor:
        '''Compute the domain adaptation loss on a labeled source
        and unlabeled target domain batch.
        
        Parameters
        ----------
        unlabeled_sample : dict
            input - torch.FloatTensor
                [BatchU, Features] minibatch of unlabeled samples.
            output - torch.LongTensor
                zeros.
        labeled_sample : dict
            input - torch.FloatTensor
                [BatchL, Features] minibatch of labeled examples.
            output - torch.LongTensor 
                one-hot labels.
        weight : float
            weight for reversed gradients passed up to the embedding
            layer. gradients used for the domain classifier are normal
            gradients, but we weight and reverse the gradients flowing 
            upward to the embedding layer by this constant.
        pseudolabel_confidence : torch.Tensor
            [BatchU,] boolean identifying observations in `unlabeled_sample`
            with confident pseudolabels.
            if not None and `self.use_conf_pseudolabels`, only performs
            domain discrimination on unlabeled samples with confident
            pseudolabels.
            
        Returns
        -------
        dan_loss : torch.FloatTensor
            domain adversarial loss term.
        '''
        
        ########################################
        # (1) Create domain labels
        ########################################
        
        # check if domain labels are provided, if not assume
        # train and target are separate domains
        # domain labels of -1 indicate `None` was passed as a domain label
        # to `SingleCellDS`
        if torch.sum(labeled_sample.get('domain', torch.Tensor([-1])) == -1) > 0:
            source_label = torch.zeros(labeled_sample['input'].size(0)).long()
            source_label = torch.nn.functional.one_hot(
                source_label, 
                num_classes=2,
            )
        else:
            # domain labels should already by one-hot
            source_label = labeled_sample['domain']
        source_label = source_label.to(device=labeled_sample['input'].device)
        
        if torch.sum(unlabeled_sample.get('domain', torch.Tensor([-1])) == -1) > 0:
            target_label = torch.ones(unlabeled_sample['input'].size(0)).long()
            target_label = torch.nn.functional.one_hot(
                target_label, 
                num_classes=2,
            )
        else:
            target_label = unlabeled_sample['domain']
        target_label = target_label.to(device=unlabeled_sample['input'].device)
        
        lx = labeled_sample['input']
        ux = unlabeled_sample['input']
        
        ########################################
        # (2) Check confidence of unlabeled obs
        ########################################
        
        if self.use_conf_pseudolabels and pseudolabel_confidence is not None:
            # check confidence of unlabeled observations and remove
            # any unconfident observations from the minibatch
            ux = ux[pseudolabel_confidence]
            target_label = target_label[pseudolabel_confidence]
            # store the number of confident unlabeled obs
        self.n_conf_pseudolabels = ux.size(0)
        self.n_total_unlabeled = unlabeled_sample['input'].size(0)
        p_conf_pseudolabels = self.n_conf_pseudolabels / self.n_total_unlabeled
        ########################################
        # (3) Embed points and Classify domains
        ########################################
        
        x = torch.cat([lx,ux], 0)
        dlabel = torch.cat([source_label, target_label], 0)
        
        self.dann.set_rev_grad_weight(weight=weight)
        domain_pred, x_embed = self.dann(x)
        
        # store embeddings and labels
        if x_embed.size(0) >= self.x_embed.size(0):
            self.x_embed = copy.copy(
                x_embed.detach().cpu()
            )
            self.dlabel  = copy.copy(
                dlabel.detach().cpu()         
            )
        
        ########################################
        # (4) Compute DAN loss
        ########################################
        
        dan_loss = self.dan_criterion(
            domain_pred,
            dlabel,
        )
        
        ########################################
        # (5) Compute DAN accuracy for logs
        ########################################
        
        _, dan_pred = torch.max(domain_pred, dim=1)
        _, dlabel_int = torch.max(dlabel, dim=1)
        self.dan_acc = torch.sum(
            dan_pred == dlabel_int,
        ) / float(dan_pred.size(0))
        
        if self.scale_loss_pseudoconf:
            dan_loss *= p_conf_pseudolabels

        return dan_loss


'''Loss weight scheduling'''


class ICLWeight(object):

    def __init__(
        self,
        ramp_epochs: int,
        burn_in_epochs: int = 0,
        max_unsup_weight: float = 10.0,
        sigmoid: bool = False,
    ) -> None:
        '''Schedules the interpolation consistency loss 
        weights across a set of epochs.

        Parameters
        ----------
        ramp_epochs : int
            number of epochs to increase the unsupervised
            loss weight until reaching a maximum value.
        burn_in_epochs : int
            epochs to wait before increasing the unsupervised loss.
        max_unsup_weight : float
            maximum weight for the unsupervised loss component.
        sigmoid : bool
            scale weight using a sigmoid function.

        Returns
        -------
        None.
        '''
        self.ramp_epochs = ramp_epochs
        self.burn_in_epochs = burn_in_epochs
        self.max_unsup_weight = max_unsup_weight
        self.sigmoid = sigmoid
        # don't allow division by zero, set step size manually
        if self.ramp_epochs == 0.:
            self.step_size = self.max_unsup_weight
        else:
            self.step_size = self.max_unsup_weight / self.ramp_epochs
        print('Scaling ICL over %d epochs, %d epochs for burn in.'
              % (self.ramp_epochs, self.burn_in_epochs))
        return
    
    def _get_weight(
        self,
        epoch: int,
    ) -> float:
        '''Compute the current weight'''
        if epoch >= (self.ramp_epochs + self.burn_in_epochs):
            weight = self.max_unsup_weight
        elif self.sigmoid:
            x = (epoch - self.burn_in_epochs) / self.ramp_epochs
            coef = np.exp(-5 * (x-1)**2)
            weight = coef * self.max_unsup_weight
        else:
            weight = self.step_size*(epoch-self.burn_in_epochs)
            
        return weight

    def __call__(
        self, 
        epoch: int,
    ) -> float:
        '''Compute the weight for an unsupervised IC loss
        given the epoch.

        Parameters
        ----------
        epoch : int
            current training epoch.

        Returns
        -------
        weight : float
            weight for the unsupervised component of IC loss.
        '''
        if type(epoch) != int:
            raise TypeError(f'epoch must be int, you passed a {type(epoch)}')
        if epoch < self.burn_in_epochs:
            weight = 0.
        else:
            weight = self._get_weight(epoch)
        return weight



import pytest
import os
import os.path as osp
import sys
import numpy as np
import torch

sys.path.append('../')


def test_dataset():
    '''Test loading samples from a dataset.
    Checks dataset creation and index parsing.
    '''
    import scanpy as sc
    from scnym.dataprep import SingleCellDS
    torch.manual_seed(1)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    # add fake domain labels
    adata.obs['domain'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
        domain = np.array(adata.obs['domain']),
    )
    
    # draw a sample
    idx = 10
    sample = ds[idx]
    sample['input'] = sample['input'].unsqueeze(0)
    # ensure the sample is the correct index
    assert np.all(
        sample['input'].numpy() == adata.X[idx, :].toarray().flatten()
    )
    # ensure the class label is correct
    y_gt = np.array(adata.obs['class'])[idx]
    y_onehot = sample['output'].view(1, -1)
    _, y_ds = torch.max(y_onehot, dim=1)
    assert np.all(y_ds.item() == y_gt)
    # ensure the domain label is correct
    d_gt = np.array(adata.obs['domain'])[idx]
    d_onehot = sample['domain'].view(1, -1)
    _, d_ds = torch.max(d_onehot, dim=1)
    assert np.all(d_ds.item() == d_gt)    
    
    # throw an error by passing an invalid idx
    
    with pytest.raises(TypeError):
        sample = ds['bad_index_type']
    
    return


def test_dataset_sparse():
    '''Test loading samples from a dataset using sparse matrices.
    Checks dataset creation and index parsing.
    '''
    import scanpy as sc
    from scnym.dataprep import SingleCellDS
    from scipy import sparse
    torch.manual_seed(1)
    np.random.seed(1)
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    X = sparse.csr_matrix(adata.X)
    
    ds = SingleCellDS(
        X = X,
        y = np.array(adata.obs['class']),
    )
    
    # draw a sample
    idx = 10
    sample = ds[idx]
    sample['input'] = sample['input'].unsqueeze(0)
    # ensure the sample is the correct index
    assert np.all(
        sample['input'].numpy() == adata.X[idx, :].toarray().flatten()
    )
    
    # throw an error by passing an invalid idx
    
    with pytest.raises(TypeError):
        sample = ds['bad_index_type']
    
    return


def test_normalization():
    '''Test log transformations and library size normalization'''
    import scanpy as sc
    from scnym.dataprep import SingleCellDS, LibrarySizeNormalize, ExpMinusOne
    torch.manual_seed(1)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    
    # draw a sample
    idx = 10
    sample = ds[idx]
    sample['input'] = sample['input'].unsqueeze(0)    
    
    # normalize the vector and log1p transform
    norm = LibrarySizeNormalize(
        log1p=True,
    )
    sample_norm = norm(sample)
    
    # exponentiate the vector
    eom = ExpMinusOne()
    sample_expm1 = eom(sample)
    
    # renormalize the vector and check validity
    sample_renorm = norm(sample_expm1)
    
    assert torch.all(
        sample_renorm['input']==sample_norm['input']
    )
    
    return
    

def test_multinomial_sampling():
    '''Test that multinomial sampling works using a large
    sample depth and a permutation test
    '''
    import scanpy as sc
    from scnym.dataprep import SingleCellDS, LibrarySizeNormalize, MultinomialSample
    torch.manual_seed(1)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    
    # draw a sample
    idx = 10
    sample = ds[idx]
    sample['input'] = sample['input'].unsqueeze(0)    
    # keep an unmodified copy of the sample
    sample_orig = sample.copy()
    
    # normalize the sample but do not log transform
    norm = LibrarySizeNormalize(
        log1p=False,
    )
    sample_norm = norm(sample)
        
    # get a multinomial vector with random depth
    mn = MultinomialSample(
        depth = (10000, 20000),
    )
    sample_mn = mn(sample_norm)
    # check that the depth is in the expected range
    d = torch.sum(sample_mn['input'], 1)
    lt_bidx = d < 10000
    gt_bidx = d > 20000
    assert torch.sum(lt_bidx) == 0.
    assert torch.sum(gt_bidx) == 0.
    
    # sample deeply from a multinomial
    mn = MultinomialSample(
        depth = (1000000, 1000001),
    )
    sample_mn = mn(sample_norm)
    # check that the depth is in the expected range
    d = torch.sum(sample_mn['input'], 1)
    lt_bidx = d < 1000000
    gt_bidx = d > 1000001
    assert torch.sum(lt_bidx) == 0.
    assert torch.sum(gt_bidx) == 0.    
    
    # sample using a depth ratio
    mn = MultinomialSample(
        depth_ratio = (0.25, 2),
    )
    sample_mn = mn(sample_norm)
    # check that the depth is in the expected range
    d = torch.sum(sample_mn['input'], 1)
    in_sizes = torch.sum(sample_norm['input'], 1)
    l_sz = in_sizes * 0.25
    h_sz = in_sizes * 2
    lt_bidx = d < l_sz
    gt_bidx = d > h_sz    
    assert torch.sum(lt_bidx) == 0.
    assert torch.sum(gt_bidx) == 0.     
    
    # normalize the sample abundance profile
    sample_mn_norm = norm(sample_mn)
    
    # get the l2 norm between the original sample and the
    # counts drawn from a multinomial
    d = torch.log1p(sample_mn_norm['input']) - torch.log1p(sample_orig['input'])
    test_norm = torch.norm(
        d, 2,
    )
    
    # permute genes to generate a null distribution of norms
    n_tests = 1000
    rand_norms = torch.zeros(n_tests,)
    for i in range(n_tests):
        ridx = torch.randperm(sample_orig['input'].size(0))
        d = torch.log1p(sample_mn_norm['input']) - torch.log1p(sample_orig['input'][ridx])
        rand_norms[i] = torch.norm(
            d, 
            2,
        )

    p = torch.sum(test_norm > rand_norms) // n_tests
    
    msg = f'p = {p} suggests something is rotten with multinomial sampling.'
    assert p < 0.05, msg
    
    return


def test_poisson_sampling():
    '''Test the Poisson sampling method.'''
    import scanpy as sc
    from scnym.dataprep import SingleCellDS, LibrarySizeNormalize, PoissonSample
    from torchvision.transforms import Compose
    torch.manual_seed(17)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    
    # draw a sample
    idx = 10
    sample = ds[idx]
    sample['input'] = sample['input'].unsqueeze(0)    
    # keep an unmodified copy of the sample
    sample_orig = sample.copy()
    
    # normalize the sample but do not log transform
    norm = LibrarySizeNormalize(
        log1p=False,
    )
    sample_norm = norm(sample)
    sample_norm_orig = sample_norm.copy()
    
    # sample using a depth ratio
    poisson = PoissonSample(depth_range=(0.5, 2.))
    sample_ps = poisson(sample_norm.copy())
    print('initial:')
    print(sample_norm_orig['input'][sample_norm_orig['input']>0][:20])
    print('sampled 00:')
    print(sample_ps['input'][sample_ps['input']>0][:20])
    
    # check that sample depths are in an expected range
    library_sizes = sample_ps['input'].sum(1)
    too_small_bidx = torch.sum(library_sizes < 4.5e5)
    too_large_bidx = torch.sum(library_sizes > 2.1e6)
    if not too_small_bidx == 0:
        raise RuntimeError(f'{too_small_bidx} cells had small libraries.')
    if not too_large_bidx == 0:
        raise RuntimeError(f'{too_large_bidx} cells had large libraries.')    

    # check that a second sample is different    
    sample_ps2 = poisson(sample_norm.copy())
    print('sampled 01:')
    print(sample_ps2['input'][sample_ps2['input']>0][:20])
    
    all_eq = torch.all(
        sample_ps['input'] == sample_ps2['input']
    )
    if all_eq:
        print('sampled 00:')
        print(sample_ps['input'][sample_ps['input']>0][:20])        
        print('sampled 01:')
        print(sample_ps2['input'][sample_ps2['input']>0][:20])        
        msg = f'The two Poisson samples were identitcal: {all_eq}'
        raise RuntimeError(msg)
        
    # test passing a minibatch for sampling
    dl = torch.utils.data.DataLoader(ds, batch_size=16)
    batch = next(iter(dl))
    
    tr = Compose(
        [
            LibrarySizeNormalize(log1p=False),
            PoissonSample(),
            LibrarySizeNormalize(log1p=True),
        ]
    )
    
    batch_sampled = tr(batch)
    
    return

def test_gene_masking():
    '''Test that gene masking augmentations zero out a
    random subset of genes'''
    import scanpy as sc
    from scnym.dataprep import SingleCellDS, GeneMasking
    torch.manual_seed(1)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    
    # draw a sample
    idx = 10
    sample = ds[idx]
    sample['input'] = sample['input'].unsqueeze(0)    
    sample_orig = sample.copy()
    
    # initiate a gene masker
    gm = GeneMasking(p_drop = 0.8, p_apply=1.0)
    
    sample_masked = gm(sample.copy())
    
    # check that we masked some genes
    n_genes_original = torch.sum(sample_orig['input'] > 0)
    n_genes_masked = torch.sum(sample_masked['input'] > 0)
    assert n_genes_masked < n_genes_original
    
    # check the masking is randomized
    sample_masked0 = gm(sample.copy())
    sample_masked1 = gm(sample.copy())
    
    bidx0 = sample_masked0['input']==0
    bidx1 = sample_masked1['input']==0
    assert ~torch.all(bidx0 & bidx1)
    
    return


def test_mixup():
    import scanpy as sc
    from scnym.dataprep import SingleCellDS, SampleMixUp
    torch.manual_seed(1)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    # draw a batch from the dataloader
    ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    dl = torch.utils.data.DataLoader(ds, batch_size=16)
    
    batch = next(iter(dl))
    
    # mixup the batch
    mixup_op = SampleMixUp(
        alpha = 0.3,
        keep_dominant_obs = True,
    )
    
    mixed_batch = mixup_op(batch)
    
    # test that mixup can be turned off with `alpha = 0.0`
    batch = next(iter(dl))
    mixup_op = SampleMixUp(
        alpha = 0.0,
        keep_dominant_obs = True,
    )
    
    mixed_batch = mixup_op(batch)
    
    assert torch.all(
        mixed_batch['input'] == batch['input']
    )
    
    return


def test_augmentation_schemes():
    '''Test each of the predefined augmentation schemes'''
    import scanpy as sc
    from scnym.dataprep import SingleCellDS, SampleMixUp, AUGMENTATION_SCHEMES
    torch.manual_seed(1)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    # draw a batch from the dataloader
    ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    dl = torch.utils.data.DataLoader(ds, batch_size=16)
    
    
    for k in AUGMENTATION_SCHEMES.keys():
        aug = AUGMENTATION_SCHEMES[k]
        batch = next(iter(dl))
        try:
            augmented = aug(batch)
        except RuntimeError:
            print(f'Augmentation scheme {k} failed.')
    return
        


def main():
    test_mixup()
    return


if __name__ == '__main__':
    main()


import numpy as np
from scipy import sparse
import os
import torch
import torch.nn.functional as F
from typing import Union
from .model import CellTypeCLF
from .dataprep import SingleCellDS
import tqdm


class Predicter(object):
    '''Predict cell types from expression data using `CellTypeCLF`.

    Attributes
    ----------
    model_weights : list
        paths to model weights for classification.
    labels : list
        str labels for output classes.
    n_cell_types : int
        number of output classes.
    n_genes : int
        number of input genes.
    models : list
        `nn.Module` for each set of weights in `.model_weights`.
    '''

    def __init__(
        self,
        model_weights: Union[str, list, tuple],
        n_genes: int = None,
        n_cell_types: int = None,
        labels: list = None,
        **kwargs,
    ) -> None:
        '''
        Predict cell types using pretrained weights for `CellTypeCLF`.

        Parameters
        ----------
        model_weights : str, list, tuple
            paths to pre-trained model weights. if more than one
            path to weights is provided, predicts using an ensemble
            of models.
        n_genes : int
            number of genes in the input frame.
        n_cell_types : int
            number of cell types in the output.
        labels : list
            string labels corresponding to each cell type output
        **kwargs passed to `model.CellTypeCLF`
        '''
        if type(model_weights) == str:
            self.model_weights = [model_weights]
        else:
            self.model_weights = model_weights
        self.labels = labels

        if n_cell_types is None:
            # get the number of output nodes from the pretrained model
            print('Assuming `n_cell_types` is the same as in the \
            pretrained model weights.')
            params = torch.load(self.model_weights[0], map_location='cpu')
            fkey = list(params.keys())[-1]
            self.n_cell_types = len(params[fkey])
        else:
            self.n_cell_types = n_cell_types

        # check that all the specified weights exist
        for weights in self.model_weights:
            if not os.path.exists(weights):
                raise FileNotFoundError()

        if n_genes is None:
            # get the number of input genes from the model weights
            print('Assuming `n_genes` is the same as in the \
            pretrained model weights.')
            params = torch.load(model_weights, map_location='cpu')
            fkey = list(params.keys())[0]
            self.n_genes = params[fkey].shape[1]
        else:
            self.n_genes = n_genes

        # Load each set of weights in `model_weights` into a model
        # to use in an ensemble prediction.
        self.models = []
        for weights in self.model_weights:
            model = CellTypeCLF(
                n_genes=self.n_genes,
                n_cell_types=self.n_cell_types,
                **kwargs,
            )
            model.load_state_dict(torch.load(weights, map_location='cpu'))

            if torch.cuda.is_available():
                model = model.cuda()

            self.models.append(model.eval())

        return

    def predict(
        self,
        X: Union[np.ndarray, sparse.csr.csr_matrix, torch.FloatTensor],
        output: str = None,
        batch_size: int=1024,
         **kwargs,
    ) -> (np.ndarray, list):
        '''
        Predict cell types given a matrix `X`.

        Parameters
        ----------
        X : np.ndarray, sparse.csr.csr_matrix, torch.FloatTensor
            [Cells, Genes]
        output : str
            additional output to include as an optional third tuple.
            ('prob', 'score').
        batch_size : int
            batch size to use for predictions.

        Returns
        -------
        predictions : np.ndarray
            [Cells,] ints of predicted class
        names : list
            [Cells,] str of predicted class names
        probabilities : np.ndarray
            [Cells, Types] probabilities (softmax outputs).

        Notes
        -----
        acceptable **kwarg for legacy compatibility --
        return_prob : bool
            return probabilities as an optional third output.
        '''
        if not X.shape[1] == self.n_genes:
            gs = (X.shape[1], self.n_genes)
            raise ValueError('%d genes in X, %d genes in model.' % gs)

        if 'return_prob' in kwargs:
            return_prob = kwargs['return_prob']
        else:
            return_prob = None
            
        if output not in ['prob', 'score'] and output is not None:
            msg = f'{output} is not a valid additional output.'
            raise ValueError(msg)

        # build a SingleCellDS so we can load cells onto the
        # GPU in batches
        ds = SingleCellDS(X=X, y=np.zeros(X.shape[0]))
        dl = torch.utils.data.DataLoader(
            ds,
            batch_size=batch_size,
        )

        # For each cell vector, compute a prediction
        # and a class probability vector.
        predictions = []
        scores = []
        probabilities = []

        # For each cell, compute predictions
        for data in tqdm.tqdm(dl, desc='Finding cell types'):
            
            X_batch = data['input']
            
            if torch.cuda.is_available():
                X_batch = X_batch.cuda()

            # take an average prediction across all models provided
            outs = []
            for model in self.models:
                out = model(X_batch)
                outs.append(out)
            outs = torch.stack(outs, dim=0)
            out = torch.mean(outs, dim=0)

            # save most likely prediction and output probabilities
            scores.append(
                out.detach().cpu().numpy()
            )
            
            _, pred = torch.max(out, 1)
            predictions.append(
                pred.detach().cpu().numpy()
            )

            probs = F.softmax(out, dim=1)
            probabilities.append(
                probs.detach().cpu().numpy()
            )

        predictions = np.concatenate(predictions, axis=0) # [Cells,]
        scores = np.concatenate(scores, axis=0)  # [Cells, Types]
        probabilities = np.concatenate(probabilities, axis=0)  # [Cells, Types]

        if self.labels is not None:
            names = []
            for i in range(len(predictions)):
                names += [self.labels[predictions[i]]]
        else:
            names = None

        # Parse the arguments to determine what to return
        # N.B. that `return_prob` here is to support legacy code
        # and may be removed in the future.
        if return_prob is True:
            return predictions, names, probabilities
        elif output is not None:
            if output == 'prob':
                return predictions, names, probabilities
            elif output == 'score':
                return predictions, names, scores
        else:
            return predictions, names



import torch
import torch.nn as nn


class ResBlock(nn.Module):
    '''Residual block.

    References
    ----------
    Deep Residual Learning for Image Recognition
    Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    arXiv:1512.03385
    '''

    def __init__(
        self,
        n_inputs: int,
        n_hidden: int,
    ) -> None:
        '''Residual block with fully-connected neural network
        layers.

        Parameters
        ----------
        n_inputs : int
            number of input dimensions.
        n_hidden : int
            number of hidden dimensions in the Residual Block.

        Returns
        -------
        None.
        '''
        super(ResBlock, self).__init__()

        self.n_inputs = n_inputs
        self.n_hidden = n_hidden

        # Build the initial projection layer
        self.linear00 = nn.Linear(self.n_inputs, self.n_hidden)
        self.norm00 = nn.BatchNorm1d(num_features=self.n_hidden)
        self.relu00 = nn.ReLU(inplace=True)

        # Map from the latent space to output space
        self.linear01 = nn.Linear(self.n_hidden, self.n_hidden)
        self.norm01 = nn.BatchNorm1d(num_features=self.n_hidden)
        self.relu01 = nn.ReLU(inplace=True)
        return

    def forward(self, x: torch.FloatTensor,
                ) -> torch.FloatTensor:
        '''Residual block forward pass.

        Parameters
        ----------
        x : torch.FloatTensor
            [Batch, self.n_inputs]

        Returns
        -------
        o : torch.FloatTensor
            [Batch, self.n_hidden]
        '''
        identity = x

        # Project input to the latent space
        o = self.norm00(self.linear00(x))
        o = self.relu00(o)

        # Project from the latent space to output space
        o = self.norm01(self.linear01(o))

        # Make this a residual connection
        # by additive identity operation
        o += identity
        return self.relu01(o)
    

class CellTypeCLF(nn.Module):
    '''Cell type classifier from expression data.

    Attributes
    ----------
    n_genes : int
        number of input genes in the model.
    n_cell_types : int
        number of output classes in the model.
    n_hidden : int
        number of hidden units in the model.
    n_layers : int
        number of hidden layers in the model.
    init_dropout : float
        dropout proportion prior to the first layer.
    residual : bool
        use residual connections.
    '''

    def __init__(
        self,
        n_genes: int,
        n_cell_types: int,
        n_hidden: int = 256,
        n_layers: int = 2,
        init_dropout: float = 0.0,
        residual: bool = False,
        batch_norm: bool = True,
        track_running_stats: bool = True,
    ) -> None:
        '''
        Cell type classifier from expression data.
        Linear layers with batch norm and dropout.

        Parameters
        ----------
        n_genes : int
            number of genes in the input
        n_cell_types : int
            number of cell types for the output
        n_hidden : int
            number of hidden unit
        n_layers : int
            number of hidden layers.
        init_dropout : float
            dropout proportion prior to the first layer.
        residual : bool
            use residual connections.
        batch_norm : bool
            use batch normalization in hidden layers.
        track_running_stats : bool
            track running statistics in batch norm layers.

        Returns
        -------
        None.
        '''
        super(CellTypeCLF, self).__init__()

        self.n_genes = n_genes
        self.n_cell_types = n_cell_types
        self.n_hidden = n_hidden
        self.n_layers = n_layers
        self.init_dropout = init_dropout
        self.residual = residual
        self.batch_norm = batch_norm
        self.track_running_stats = track_running_stats
            
        # simulate technical dropout of scRNAseq
        self.init_dropout = nn.Dropout(p=self.init_dropout)

        # Define a vanilla NN layer with batch norm, dropout, ReLU
        vanilla_layer = [
            nn.Linear(self.n_hidden, self.n_hidden),
        ]
        if self.batch_norm:
            vanilla_layer += [
                nn.BatchNorm1d(
                    num_features=self.n_hidden, 
                    track_running_stats=self.track_running_stats,
                ),
            ]
        vanilla_layer += [
            nn.Dropout(),
            nn.ReLU(inplace=True),
        ]

        # Define a residual NN layer with batch norm, dropout, ReLU
        residual_layer = [
            ResBlock(self.n_hidden, self.n_hidden),
        ]
        if self.batch_norm:
            residual_layer += [
                nn.BatchNorm1d(
                    num_features=self.n_hidden,
                    track_running_stats=self.track_running_stats,                    
                ),         
            ]
        
        residual_layer += [
            nn.Dropout(),
            nn.ReLU(inplace=True),
        ]

        # Build the intermediary layers of the model
        if self.residual:
            hidden_layer = residual_layer
        else:
            hidden_layer = vanilla_layer

        hidden_layers = hidden_layer*self.n_layers

        # Build the classifier `nn.Module`.
        self.classif = nn.Sequential(
            nn.Linear(self.n_genes, self.n_hidden),
            nn.BatchNorm1d(
                num_features=self.n_hidden,
                track_running_stats=self.track_running_stats,                
            ),
            nn.Dropout(),
            nn.ReLU(inplace=True),
            *hidden_layers,
            nn.Linear(self.n_hidden, self.n_cell_types)
        )
        return

    def forward(
        self,
        x: torch.FloatTensor,
    ) -> torch.FloatTensor:
        '''Perform a forward pass through the model

        Parameters
        ----------
        x : torch.FloatTensor
            [Batch, self.n_genes]

        Returns
        -------
        pred : torch.FloatTensor
            [Batch, self.n_cell_types]
        '''
        # add initial dropout noise
        if self.init_dropout.p > 0:
            # expm1 to normed counts
            x = torch.expm1(x)
            x = self.init_dropout(x)
            # renorm to log1p CPM
            size = torch.sum(x, dim=1).reshape(-1, 1)
            prop_input_ = x / size
            norm_input_ = prop_input_ * 1e6
            x = torch.log1p(norm_input_)
        pred = self.classif(x)
        return pred
    
    
class GradReverse(torch.autograd.Function):
    '''Layer that reverses and scales gradients before
    passing them up to earlier ops in the computation graph
    during backpropogation.
    '''
    @staticmethod
    def forward(ctx, x, weight):
        '''
        Perform a no-op forward pass that stores a weight for later
        gradient scaling during backprop.
        
        Parameters
        ----------
        x : torch.FloatTensor
            [Batch, Features]
        weight : float
            weight for scaling gradients during backpropogation.
            stored in the "context" ctx variable.
        
        Notes
        -----
        We subclass `Function` and use only @staticmethod as specified
        in the newstyle pytorch autograd functions.
        https://pytorch.org/docs/stable/autograd.html#torch.autograd.Function
        
        We define a "context" ctx of the class that will hold any values
        passed during forward for use in the backward pass.
        
        `x.view_as(x)` and `*1` are necessary so that `GradReverse`
        is actually called
        `torch.autograd` tries to optimize backprop and 
        excludes no-ops, so we have to trick it :)
        '''
        # store the weight we'll use in backward in the context
        ctx.weight = weight
        return x.view_as(x) * 1
    
    @staticmethod
    def backward(ctx, grad_output):
        '''Return gradients
        
        Returns
        -------
        rev_grad : torch.FloatTensor
            reversed gradients scaled by `weight` passed in `.forward()`
        None : None
            a dummy "gradient" required since we passed a weight float
            in `.forward()`.
        '''
        # here scale the gradient and multiply by -1
        # to reverse the gradients
        return (grad_output * -1 * ctx.weight), None

    
class DANN(nn.Module):
    '''Build a domain adaptation neural network'''
    
    def __init__(
        self,
        model: CellTypeCLF,
        n_domains: int=2,
        weight: float=1.,
        n_layers: int=1,
    ) -> None:
        '''Build a domain adaptation neural network using
        the embedding of a provided model.
        
        Parameters
        ----------
        model : CellTypeCLF
            cell type classification model.
        n_domains : int
            number of domains to adapt.
        weight : float
            weight for reversed gradients.
        n_layers : int
            number of hidden layers in the network.
        
        Returns
        -------
        None.
        '''
        super(DANN, self).__init__()
        
        self.model = model
        self.n_domains = n_domains
        
        self.embed = nn.Sequential(
            *list(list(model.modules())[0].children())[1][:-1]
        )
        
        hidden_layers = [
            nn.Linear(self.model.n_hidden, self.model.n_hidden),
            nn.ReLU(),
        ] * n_layers
        
        self.domain_clf = nn.Sequential(
            *hidden_layers,
            nn.Linear(self.model.n_hidden, self.n_domains),
        )
        return
    
    def set_rev_grad_weight(
        self,
        weight: float,
    ) -> None:
        '''Set the weight term used after reversing gradients'''
        self.weight = weight
        return
    
    def forward(
        self,
        x: torch.FloatTensor,
    ) -> (torch.FloatTensor, torch.FloatTensor):
        '''Perform a forward pass.
        
        Parameters
        ----------
        x : torch.FloatTensor
            [Batch, Features] input.
        
        Returns
        -------
        domain_pred : torch.FloatTensor
            [Batch, n_domains] logits.
        x_embed : torch.FloatTensor
            [Batch, n_hidden]
        '''
        # get the model embedding
        x_embed = self.embed(x)
        # reverse gradients and scale by a weight
        x_rev   = GradReverse.apply(
            x_embed,
            self.weight,
        )
        # classify the domains
        domain_pred = self.domain_clf(x_rev)
        return domain_pred, x_embed
        


class CellTypeCLFConditional(CellTypeCLF):
    '''Conditional vartiaton of the `CellTypeCLF`

    Attributes
    ----------
    n_genes : int
        number of the input features corresponding to genes.
    n_tissues : int
        length of the one-hot `upper_group` vector appended
        to inputs.
    '''

    def __init__(
        self,
        n_genes: int,
        n_tissues: int,
        **kwargs,
    ) -> None:
        '''Conditional vartiaton of the `CellTypeCLF`.

        Parameters
        ----------
        n_genes : int
            number of genes in the input
        n_tissues : int
            number of tissues encoded in the conditional vector.

        Returns
        -------
        None.

        Notes
        -----
        Assumes that inputs are `n_genes + n_tissues`, concatenated
        [Genes :: Tissue-One-Hot].

        Passes `**kwargs` to `CellTypeCLF`.
        '''
        # Build a CellTypeCLF with `n_genes` + `n_tissues` input nodes to
        # take both the gene vector and one-hot upper_group label as input
        super(CellTypeCLFConditional, self).__init__(
            n_genes=(n_genes + n_tissues),
            **kwargs)
        self.n_tissues = n_tissues
        return

    def forward(
        self, 
        x: torch.FloatTensor,
    ) -> torch.FloatTensor:
        '''Perform a forward pass through the model

        Parameters
        ----------
        x : torch.FloatTensor
            [Batch, self.n_genes + self.n_tissues]

        Returns
        -------
        pred : torch.FloatTensor
            [Batch, self.n_cell_types]
        '''
        # don't pass one_hot labels through the initial dropout
        one_hot = x[:, -self.n_tissues:]
        genes = x[:, :-self.n_tissues]

        genes = self.init_dropout(genes)
        x_drop = torch.cat([genes, one_hot], dim=1)

        # classify on the full genes + one-hot input
        pred = self.classif(x_drop)
        return pred



'''
Utility functions
'''
import torch
import numpy as np
import anndata
from scipy import sparse
import pandas as pd
import tqdm
from scipy import stats
import scanpy as sc
from sklearn.neighbors import NearestNeighbors, KNeighborsRegressor
from sklearn.metrics.pairwise import euclidean_distances
from typing import Union, Callable


def make_one_hot(
    labels: torch.LongTensor, 
    C=2,
) -> torch.FloatTensor:
    '''
    Converts an integer label torch.autograd.Variable to a one-hot Variable.

    Parameters
    ----------
    labels : torch.LongTensor or torch.cuda.LongTensor
        [N, 1], where N is batch size.
        Each value is an integer representing correct classification.
    C : int
        number of classes in labels.

    Returns
    -------
    target : torch.FloatTensor or torch.cuda.FloatTensor
        [N, C,], where C is class number. One-hot encoded.
    '''
    if labels.ndimension() < 2:
        labels = labels.unsqueeze(1)
    one_hot = torch.zeros([labels.size(0), C, ],
                          dtype=torch.float32, device=labels.device)
    target = one_hot.scatter_(1, labels, 1)

    return target


def l1_layer0(
    model: torch.nn.Module,
) -> torch.FloatTensor:
    '''Compute l1 norm for the first input layer of
    a `CellTypeCLF` model.

    Parameters
    ----------
    model : torch.nn.Module
        CellTypeCLF model with `.classif` module.

    Returns
    -------
    l1_reg : torch.FloatTensor
        [1,] l1 norm for the first layer parameters.
    '''
    # get the parameters of the first classification layer
    layer0 = list(model.classif.modules())[1]
    params = layer0.parameters()
    l1_reg = None

    # compute the l1_norm
    for W in params:
        if l1_reg is None:
            l1_reg = W.norm(1)
        else:
            l1_reg = l1_reg + W.norm(1)
    return l1_reg


def append_categorical_to_data(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    categorical: np.ndarray,
) -> (Union[np.ndarray, sparse.csr.csr_matrix], np.ndarray):
    '''Convert `categorical` to a one-hot vector and append
    this vector to each sample in `X`.

    Parameters
    ----------
    X : np.ndarray, sparse.csr.csr_matrix
        [Cells, Features]
    categorical : np.ndarray
        [Cells,]

    Returns
    -------
    Xa : np.ndarray
        [Cells, Features + N_Categories]
    categories : np.ndarray
        [N_Categories,] str category descriptors.
    '''
    # `pd.Categorical(xyz).codes` are int values for each unique
    # level in the vector `xyz`
    labels = pd.Categorical(categorical)
    idx = np.array(labels.codes)
    idx = torch.from_numpy(idx.astype('int32')).long()
    categories = np.array(labels.categories)

    one_hot_mat = make_one_hot(
        idx,
        C=len(categories),
    )
    one_hot_mat = one_hot_mat.numpy()
    assert X.shape[0] == one_hot_mat.shape[0], \
        'dims unequal at %d, %d' % (X.shape[0], one_hot_mat.shape[0])
    # append one hot vector to the [Cells, Features] matrix
    if sparse.issparse(X):
        X = sparse.hstack([X, one_hot_mat])
    else:
        X = np.concatenate([X, one_hot_mat], axis=1)
    return X, categories


def get_adata_asarray(
    adata: anndata.AnnData,
) -> Union[np.ndarray, sparse.csr.csr_matrix]:
    '''Get the gene expression matrix `.X` of an
    AnnData object as an array rather than a view.
    
    Parameters
    ----------
    adata : anndata.AnnData
        [Cells, Genes] AnnData experiment.
    
    Returns
    -------
    X : np.ndarray, sparse.csr.csr_matrix
        [Cells, Genes] `.X` attribute as an array
        in memory.
    
    Notes
    -----
    Returned `X` will match the type of `adata.X` view.
    '''
    if sparse.issparse(adata.X):
        X = sparse.csr.csr_matrix(adata.X)
    else:
        X = np.array(adata.X)
    return X


def build_classification_matrix(
    X: Union[np.ndarray, sparse.csr.csr_matrix],
    model_genes: np.ndarray,
    sample_genes: np.ndarray,
    gene_batch_size: int=512,
) -> Union[np.ndarray, sparse.csr.csr_matrix]:
    '''
    Build a matrix for classification using only genes that overlap
    between the current sample and the pre-trained model.

    Parameters
    ----------
    X : np.ndarray, sparse.csr_matrix
        [Cells, Genes] count matrix.
    model_genes : np.ndarray
        gene identifiers in the order expected by the model.
    sample_genes : np.ndarray
        gene identifiers for the current sample.
    gene_batch_size : int
        number of genes to copy between arrays per batch.
        controls a speed vs. memory trade-off.

    Returns
    -------
    N : np.ndarray, sparse.csr_matrix
        [Cells, len(model_genes)] count matrix.
        Values where a model gene was not present in the sample are left
        as zeros. `type(N)` will match `type(X)`.
    '''
    # check types
    if type(X) not in (np.ndarray, sparse.csr.csr_matrix):
        msg = f'X is type {type(X)}, must `np.ndarray` or `sparse.csr_matrix`'
        raise TypeError(msg)
    n_cells = X.shape[0]
    # check if gene names already match exactly
    if len(model_genes) == len(sample_genes):
        if np.all(model_genes == sample_genes):
            print('Gene names match exactly, returning input.')
            return X
    
    # instantiate a new [Cells, model_genes] matrix where columns
    # retain the order used during training
    if type(X) == np.ndarray:
        N = np.zeros((n_cells, len(model_genes)))
    else:
        # use sparse matrices if the input is sparse
        N = sparse.lil_matrix((n_cells, len(model_genes),))

    # map gene indices from the model to the sample genes
    model_genes_indices = []
    sample_genes_indices = []
    common_genes = 0
    for i, g in tqdm.tqdm(enumerate(sample_genes), desc='mapping genes'):
        if np.sum(g==model_genes) > 0:
            model_genes_indices.append(
                int(np.where(g==model_genes)[0])
            )
            sample_genes_indices.append(
                i,
            )
            common_genes += 1
            
    # copy the data in batches to the new array to avoid memory overflows
    gene_idx = 0
    n_batches = int(np.ceil(N.shape[1] / gene_batch_size))
    for b in tqdm.tqdm(range(n_batches), desc='copying gene batches'):
        model_batch_idx = model_genes_indices[gene_idx:gene_idx+gene_batch_size]
        sample_batch_idx = sample_genes_indices[gene_idx:gene_idx+gene_batch_size]
        N[:, model_batch_idx] = X[:, sample_batch_idx]
        gene_idx += gene_batch_size
            
    if sparse.issparse(N):
        # convert to `csr` from `csc`
        N = sparse.csr_matrix(N)
    print('Found %d common genes.' % common_genes)
    return N


def knn_smooth_pred_class(
    X: np.ndarray,
    pred_class: np.ndarray,
    grouping: np.ndarray = None,
    k: int = 15,
) -> np.ndarray:
    '''
    Smooths class predictions by taking the modal class from each cell's
    nearest neighbors.

    Parameters
    ----------
    X : np.ndarray
        [N, Features] embedding space for calculation of nearest neighbors.
    pred_class : np.ndarray
        [N,] array of unique class labels.
    groupings : np.ndarray
        [N,] unique grouping labels for i.e. clusters.
        if provided, only considers nearest neighbors *within the cluster*.
    k : int
        number of nearest neighbors to use for smoothing.

    Returns
    -------
    smooth_pred_class : np.ndarray
        [N,] unique class labels, smoothed by kNN.

    Examples
    --------
    >>> smooth_pred_class = knn_smooth_pred_class(
    ...     X = X,
    ...     pred_class = raw_predicted_classes,
    ...     grouping = louvain_cluster_groups,
    ...     k = 15,)

    Notes
    -----
    scNym classifiers do not incorporate neighborhood information.
    By using a simple kNN smoothing heuristic, we can leverage neighborhood
    information to improve classification performance, smoothing out cells
    that have an outlier prediction relative to their local neighborhood.
    '''
    if grouping is None:
        # do not use a grouping to restrict local neighborhood
        # associations, create a universal pseudogroup `0`.
        grouping = np.zeros(X.shape[0])

    smooth_pred_class = np.zeros_like(pred_class)
    for group in np.unique(grouping):
        # identify only cells in the relevant group
        group_idx = np.where(grouping == group)[0].astype('int')
        X_group = X[grouping == group, :]
        # if there are < k cells in the group, change `k` to the
        # group size
        if X_group.shape[0] < k:
            k_use = X_group.shape[0]
        else:
            k_use = k
        # compute a nearest neighbor graph and identify kNN
        nns = NearestNeighbors(n_neighbors=k_use,).fit(X_group)
        dist, idx = nns.kneighbors(X_group)

        # for each cell in the group, assign a class as
        # the majority class of the kNN
        for i in range(X_group.shape[0]):
            classes = pred_class[group_idx[idx[i, :]]]
            uniq_classes, counts = np.unique(classes, return_counts=True)
            maj_class = uniq_classes[int(np.argmax(counts))]
            smooth_pred_class[group_idx[i]] = maj_class
    return smooth_pred_class



class RBFWeight(object):
    
    def __init__(
        self,
        alpha: float=None,
    ) -> None:
        '''Generate a set of weights based on distances to a point
        with a radial basis function kernel.
        
        Parameters
        ----------
        alpha : float
            radial basis function parameter. inverse of sigma
            for a standard Gaussian pdf.
        
        Returns
        -------
        None.
        '''
        self.alpha = alpha
        return
    
    def set_alpha(
        self,
        X: np.ndarray,
        n_max: int=None,
        dm: np.ndarray=None,
    ) -> None:
        '''Set the alpha parameter of a Gaussian RBF kernel
        as the median distance between points in an array of
        observations.
        
        Parameters
        ----------
        X : np.ndarray
            [N, P] matrix of observations and features.
        n_max : int
            maximum number of observations to use for median
            distance computation.
        dm : np.ndarray, optional
            [N, N] distance matrix for setting the RBF kernel parameter.
            speeds computation if pre-computed.            
        
        Returns
        -------
        None. Sets `self.alpha`.
    
        References
        ----------
        A Kernel Two-Sample Test
        Arthur Gretton, Karsten M. Borgwardt, Malte J. Rasch, 
        Bernhard Schölkopf, Alexander Smola.
        JMLR, 13(Mar):723−773, 2012.
        http://jmlr.csail.mit.edu/papers/v13/gretton12a.html        
        '''
        if n_max is None:
            n_max = X.shape[0]

        if dm is None:
            # compute a distance matrix from observations
            if X.shape[0] > n_max:
                ridx = np.random.choice(
                    X.shape[0], 
                    size=n_max, 
                    replace=False,
                )
                X_p = X[ridx, :]
            else:
                X_p = X

            dm = euclidean_distances(X_p,)
        
        upper = dm[np.triu_indices_from(dm, k=1)]

        # overwrite_input = True saves memory by overwriting
        # the upper indices in the distance matrix array during
        # median computation
        sigma = np.median(
            upper,
            overwrite_input=True,
        )
        self.alpha = 1./(2*(sigma**2))
        return

    def __call__(
        self,
        distances: np.ndarray,
    ) -> np.ndarray:
        '''Generate a set of weights based on distances to a point
        with a radial basis function kernel.

        Parameters
        ----------
        distances : np.ndarray
            [N,] distances used to generate weights.

        Returns
        -------
        weights : np.ndarray
            [N,] weights from the radial basis function kernel.

        Notes
        -----
        We weight distances with a Gaussian RBF.

        .. math::

            f(r) = \exp -(\alpha r)^2

        '''
        # check that alpha parameter is set
        if self.alpha is None:
            msg = 'must set `alpha` attribute before computing weights.\n'
            msg += 'use `.set_alpha() method to estimate from data.'
            raise ValueError(msg)
        
        # generate weights with an RBF kernel
        weights = np.exp(-(self.alpha * distances)**2)
        return weights


def knn_smooth_pred_class_prob(
    X: np.ndarray,
    pred_probs: np.ndarray,
    names: np.ndarray,    
    grouping: np.ndarray = None,
    k: Union[Callable,int] = 15,
    dm: np.ndarray=None,
    **kwargs,
) -> np.ndarray:
    '''
    Smooths class predictions by taking the modal class from each cell's
    nearest neighbors.

    Parameters
    ----------
    X : np.ndarray
        [N, Features] embedding space for calculation of nearest neighbors.
    pred_probs : np.ndarray
        [N, C] array of class prediction probabilities.
    names : np.ndarray,
        [C,] names of predicted classes in `pred_probs`.
    groupings : np.ndarray
        [N,] unique grouping labels for i.e. clusters.
        if provided, only considers nearest neighbors *within the cluster*.
    k : int
        number of nearest neighbors to use for smoothing.
    dm : np.ndarray, optional
        [N, N] distance matrix for setting the RBF kernel parameter.
        speeds computation if pre-computed.

    Returns
    -------
    smooth_pred_class : np.ndarray
        [N,] unique class labels, smoothed by kNN.

    Examples
    --------
    >>> smooth_pred_class = knn_smooth_pred_class_prob(
    ...     X = X,
    ...     pred_probs = predicted_class_probs,
    ...     grouping = louvain_cluster_groups,
    ...     k = 15,)

    Notes
    -----
    scNym classifiers do not incorporate neighborhood information.
    By using a simple kNN smoothing heuristic, we can leverage neighborhood
    information to improve classification performance, smoothing out cells
    that have an outlier prediction relative to their local neighborhood.
    '''
    if grouping is None:
        # do not use a grouping to restrict local neighborhood
        # associations, create a universal pseudogroup `0`.
        grouping = np.zeros(X.shape[0])

    smooth_pred_probs = np.zeros_like(pred_probs)        
    smooth_pred_class = np.zeros(pred_probs.shape[0], dtype='object')
    for group in np.unique(grouping):
        # identify only cells in the relevant group
        group_idx = np.where(grouping == group)[0].astype('int')
        X_group = X[grouping == group, :]
        y_group = pred_probs[grouping == group, :]
        # if k is a Callable, use it to define k for this group
        if callable(k):
            k_use = k(X_group.shape[0])
        else:
            k_use = k
        
        # if there are < k cells in the group, change `k` to the
        # group size
        if X_group.shape[0] < k_use:
            k_use = X_group.shape[0]

        # set up weights using a radial basis function kernel
        rbf = RBFWeight()
        rbf.set_alpha(
            X=X_group,
            n_max=None,
            dm=dm,
        )
        
        if 'dm' in kwargs:
            del kwargs['dm']
        # fit a nearest neighbor regressor
        nns = KNeighborsRegressor(
            n_neighbors=k_use,
            weights=rbf,
            **kwargs,
        ).fit(X_group, y_group)
        smoothed_probs = nns.predict(X_group)

        smooth_pred_probs[group_idx, :] = smoothed_probs
        g_classes = names[np.argmax(smoothed_probs, axis=1)]
        smooth_pred_class[group_idx] = g_classes
        
    return smooth_pred_class


def argmax_pred_class(grouping: np.ndarray,
                      prediction: np.ndarray,
                      ):
    '''Assign class to elements in groups based on the
    most common predicted class for that group.

    Parameters
    ----------
    grouping : np.ndarray
        [N,] partition values defining groups to be classified.
    prediction : np.ndarray
        [N,] predicted values for each element in `grouping`.

    Returns
    -------
    assigned_classes : np.ndarray
        [N,] class labels based on the most common class assigned
        to elements in the group partition.

    Examples
    --------
    >>> grouping = np.array([0,0,0,1,1,1,2,2,2,2])
    >>> prediction = np.array(['A','A','A','B','A','B','C','A','B','C'])
    >>> argmax_pred_class(grouping, prediction)
    np.ndarray(['A','A','A','B','B','B','C','C','C','C',])

    Notes
    -----
    scNym classifiers do not incorporate neighborhood information.
    This simple heuristic leverages cluster information obtained by
    an orthogonal method and assigns all cells in a given cluster
    the majority class label within that cluster.
    '''
    assert grouping.shape[0] == prediction.shape[0], \
        '`grouping` and `prediction` must be the same length'
    groups = sorted(list(set(grouping.tolist())))

    assigned_classes = np.zeros(grouping.shape[0], dtype='object')

    for i, group in enumerate(groups):
        classes, counts = np.unique(prediction[grouping == group],
                                    return_counts=True)
        majority_class = classes[np.argmax(counts)]
        assigned_classes[grouping == group] = majority_class
    return assigned_classes


def compute_entropy_of_mixing(
    X: np.ndarray,
    y: np.ndarray,
    n_neighbors: int,
    n_iters: int=None,
    **kwargs,
) -> np.ndarray:
    '''Compute the entropy of mixing among groups given
    a distance matrix.
    
    Parameters
    ----------
    X : np.ndarray
        [N, P] feature matrix.
    y : np.ndarray
        [N,] group labels.
    n_neighbors : int
        number of nearest neighbors to draw for each iteration 
        of the entropy computation.
    n_iters : int
        number of iterations to perform.
        if `n_iters is None`, uses every point.
        
    Returns
    -------
    entropy_of_mixing : np.ndarray
        [n_iters,] entropy values for each iteration.

    Notes
    -----
    The entropy of batch mixing is computed by sampling `n_per_sample` 
    cells from a local neighborhood in the nearest neighbor graph
    and contructing a probability vector based on their group membership.
    The entropy of this probability vector is computed as a metric of
    intermixing between groups.
    
    If groups are more mixed, the probability vector will have higher
    entropy, and vice-versa.
    '''
    # build nearest neighbor graph
    n_neighbors = min(n_neighbors, X.shape[0])
    nn = NearestNeighbors(
        n_neighbors=n_neighbors,
        metric='euclidean',
        **kwargs,
    )
    nn.fit(X)
    nn_idx = nn.kneighbors(return_distance=False)
    
    # define query points
    if n_iters is not None:
        # don't duplicate points when sampling
        n_iters = min(n_iters, X.shape[0])
    
    if (n_iters is None) or (n_iters == X.shape[0]):
        # sample all points
        query_points = np.arange(X.shape[0])
    else:
        # subset random query points for entropy
        # computation
        assert n_iters < X.shape[0]
        query_points = np.random.choice(
            X.shape[0],
            size=n_iters,
            replace=False,
        )
    
    entropy_of_mixing = np.zeros(len(query_points))
    for i, ridx in enumerate(query_points): 
        # get the nearest neighbors of a point
        nn_y = y[nn_idx[ridx, :]]
        
        nn_y_p = np.zeros(len(np.unique(y)))
        for j, v in enumerate(np.unique(y)):
            nn_y_p[j] = sum(nn_y == v)
        nn_y_p = nn_y_p / nn_y_p.sum()
        
        # use base 2 to return values in bits rather
        # than the default nats
        H = stats.entropy(nn_y_p)
        entropy_of_mixing[i] = H
    return entropy_of_mixing


'''Find new cell state based on scNym confidence scores'''

from sklearn.metrics import calinski_harabasz_score
def _optimize_clustering(adata, resolution: list=[0.1, 0.2, 0.3, 0.5, 1.0]):
    scores = []
    for r in resolution:
        sc.tl.leiden(adata, resolution=r)
        s = calinski_harabasz_score(adata.obsm['X_scnym'], adata.obs['leiden'])
        scores.append(s)
    cl_opt_df = pd.DataFrame({'resolution': resolution, 'score': scores})
    best_idx = np.argmax(cl_opt_df['score'])
    res = cl_opt_df.iloc[best_idx, 0]
    sc.tl.leiden(adata, resolution=res)
    print('Best resolution: ', res)
    return cl_opt_df


def find_low_confidence_cells(
    adata: anndata.AnnData,
    confidence_threshold: float=0.5,
    confidence_key: str='Confidence',
    use_rep: str='X_scnym',
    n_neighbors: int=15,
) -> pd.DataFrame:
    '''Find cells with low confidence predictions and suggest a potential
    number of cell states within the low confidence cell population.
    
    Parameters
    ----------
    adata : anndata.AnnData
        [Cells, Genes] experiment containing an scNym embedding and scNym
        confidence scores.
    confidence_threshold : float
        threshold for low confidence cells.
    confidence_key : str
        key in `adata.obs` containing confidence scores.
    use_rep : str
        tensor in `adata.obsm` containing the scNym embedding.
    n_neighbors : int
        number of nearest neighbors to use for NN graph construction
        prior to community detection.
    
    Returns
    -------
    None.
    Adds `adata.uns["scNym_low_confidence_cells"]`, a `dict` containing
    keys `"cluster_optimization", "n_clusters", "embedding"`.
    Adds key to `adata.obs["scNym_low_confidence_cluster"]`.
    
    Notes
    -----
    '''
    # identify low confidence cells
    adata.obs['scNym Discovery'] = (
        adata.obs[confidence_key] < confidence_threshold
    ).astype(bool)
    low_conf_bidx = adata.obs['scNym Discovery']
    
    # embed low confidence cells
    lc_ad = adata[adata.obs['scNym Discovery'], :].copy()
    sc.pp.neighbors(lc_ad, use_rep=use_rep, n_neighbors=n_neighbors)
    sc.tl.umap(lc_ad, min_dist=0.3)
    
    cl_opt_df = _optimize_clustering(lc_ad)
    
    lc_embed = lc_ad.obs.copy()
    for k in range(1, 3):
        lc_embed[f'UMAP{k}'] = lc_ad.obsm['X_umap'][:, k-1]
    
    # set the outputs
    adata.uns['scNym_low_confidence_cells'] = {
        'cluster_optimization' : cl_opt_df,
        'n_clusters' : len(np.unique(lc_ad.obs['leiden'])),
        'embedding': lc_embed,
    }
    adata.obs['scNym_low_confidence_cluster'] = 'High Confidence'
    adata.obs.loc[
        low_conf_bidx,
        'scNym_low_confidence_cluster',
    ] = lc_ad.obs['leiden'].apply(lambda x : f'Low Confidence {x}')
    return


import pytest
import os
import os.path as osp
import sys
import numpy as np
import torch
import anndata
import scipy.sparse as sparse

sys.path.append('../')

def test_build_classification_matrix_dense():
    '''Builds a classification matrix from two
    matrices with gene names that are not aligned.
    
    Tests using dense matrices.
    '''
    import scnym
    
    B = np.random.random((100, 10))
    
    # create dummy gene names where the order of 
    # genes in `B` is permuted
    A_genes = np.arange(10)
    B_genes = np.random.permutation(A_genes)
    
    # build the classification matrix
    X = scnym.utils.build_classification_matrix(
        X=B,
        model_genes=A_genes,
        sample_genes=B_genes,
    )
    
    # X should have the genes of B in the order of A
    for i, g in enumerate(A_genes):
        j = int(np.where(B_genes == g)[0])
        assert np.all(X[:, i] == B[:, j])
    return
    

def test_build_classification_matrix_sparse():
    '''Builds a classification matrix from two
    matrices with gene names that are not aligned.
    
    Tests using sparse matrices.
    '''
    import scnym
    
    # generate a sparse matrix with ~10% of elements filled
    B = np.zeros((100, 10))
    ridx = np.random.choice(B.size, size=100, replace=True)
    B.flat[ridx] = 1
    B = sparse.csr_matrix(B)
    
    # create dummy gene names where the order of 
    # genes in `B` is permuted
    A_genes = np.arange(10)
    B_genes = np.random.permutation(A_genes)
    
    # build the classification matrix
    X = scnym.utils.build_classification_matrix(
        X=B,
        model_genes=A_genes,
        sample_genes=B_genes,
    )
    assert sparse.issparse(X)
    
    # X should have the genes of B in the order of A
    for i, g in enumerate(A_genes):
        j = int(np.where(B_genes == g)[0])
        assert np.all(X[:, i].toarray() == B[:, j].toarray())
    return


def test_get_adata_asarray():
    '''Tests getting anndata main matrix as an array in memory'''
    # test getting a dense matrix
    import scnym
    
    adata = anndata.AnnData(
        X = np.random.random((100, 100))
    )
    X = scnym.utils.get_adata_asarray(adata=adata)
    assert type(X) == np.ndarray
    
    # test getting a sparse matrix
    A = np.zeros((100, 100))
    ridx = np.random.choice(A.size, size=1000, replace=True)
    A.flat[ridx] = 1
    A = sparse.csr_matrix(A)
    adata = anndata.AnnData(
        X = A
    )
    X = scnym.utils.get_adata_asarray(adata=adata)
    assert sparse.issparse(X)
    return


import pytest
import numpy as np
import torch
import anndata
import os
import os.path as osp
from pathlib import Path
import scanpy as sc
from sklearn.model_selection import StratifiedKFold

import sys
sys.path = ['../'] + sys.path
import scnym
scnym_api = scnym.api.scnym_api
atlas2target = scnym.api.atlas2target
scnym_tune = scnym.api.scnym_tune
split_data = scnym.api.split_data
TEST_URL = scnym.api.TEST_URL


def cuda_only(func):
    '''Decorator that sets a test to run only if CUDA is available'''
    def wrap_cuda_only():
        if not torch.cuda.is_available():
            print('Test only run if CUDA compute device is available.')
            return
        else:
            func()
    return wrap_cuda_only


def test_scnym():
    '''Test scNym training and prediction'''
    np.random.seed(1)
    # load a testing dataset
    adata = sc.datasets._datasets.read(sc.settings.datasetdir / 'kang17.h5ad', backup_url=TEST_URL)
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['cell'] = np.array(adata.obs['cell'])
    adata.obs['ground_truth'] = np.array(adata.obs['cell'])
    adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    # downsample to speed up testing
    ridx = np.random.choice(
        adata.shape[0],
        size=2048,
        replace=False,
    )
    adata = adata[ridx, :].copy()
    
    # train an scNym model
    config = {'n_epochs': 1}
    scnym_api(
        adata=adata,
        task='train',
        groupby='cell',
        out_path=str(sc.settings.datasetdir),
        config=config,
    )
    
    assert 'scNym_train_results' in adata.uns.keys()
    assert osp.exists(osp.join(str(sc.settings.datasetdir), '00_best_model_weights.pkl'))
    
    # predict cell types with an scNym model
    scnym_api(
        adata=adata,
        task='predict',
        key_added='scNym',
        out_path=str(sc.settings.datasetdir),
        trained_model=str(sc.settings.datasetdir),
        config=config,
    )
    
    assert 'X_scnym' in adata.obsm.keys()
    assert 'scNym' in adata.obs.columns
    
    # check accuracy
    target_gt = np.array(
        adata.obs.loc[target_bidx, 'ground_truth']
    )
    pred = np.array(adata.obs.loc[target_bidx, 'scNym'])
    print('Example Truth vs. Prediction')
    for i in range(10):
        print(f'{target_gt[i]}\t|\t{pred[i]}')
    print()
    acc = np.sum(target_gt == pred) / len(pred) * 100
    print(f'Accuracy on target set: {acc}%')
    
    return


def test_assumption_checking():
    '''Test that errors are thrown if assumptions are violated'''
    np.random.seed(1)
    # load a testing dataset
    adata = sc.datasets._datasets.read(sc.settings.datasetdir / 'kang17.h5ad', backup_url=TEST_URL)
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['cell'] = np.array(adata.obs['cell'])
    adata.obs['ground_truth'] = np.array(adata.obs['cell'])
    adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    # downsample to speed up testing
    ridx = np.random.choice(
        adata.shape[0],
        size=2048,
        replace=False,
    )
    adata = adata[ridx, :].copy()
    
    
    # test that an input anndata with duplicate genes
    # throws an error
    adata_dup_genes = adata.copy()
    var_names_with_dups = np.array(adata_dup_genes.var_names)
    var_names_with_dups[-1] = var_names_with_dups[-2]
    adata_dup_genes.var_names = var_names_with_dups
    
    with pytest.raises(ValueError, match='Duplicate Genes'):
        # this should throw an error about duplicate genes
        config = {'n_epochs': 1}
        scnym_api(
            adata=adata_dup_genes,
            task='train',
            groupby='cell',
            out_path=str(sc.settings.datasetdir),
            config=config,
        )
        
    # test that an input anndata with `.X` formatted as something
    # other than log1p(CPM) will throw an error
    adata_not_log1p = adata.copy()
    adata_not_log1p.X = np.expm1(adata_not_log1p.X)

    with pytest.raises(ValueError, match='Normalization'):
        # this should throw an error with instructions on how to 
        # normalize data
        scnym_api(
            adata=adata_not_log1p,
            task='train',
            groupby='cell',
            out_path=str(sc.settings.datasetdir),
            config=config,
        )
    return


def test_pretrained():
    '''Test that pretrained models are disabled'''
    np.random.seed(1)    
    # load a testing dataset
    adata = sc.datasets._datasets.read(sc.settings.datasetdir / 'kang17.h5ad', backup_url=TEST_URL)
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['cell'] = np.array(adata.obs['cell'])
    adata.obs['ground_truth'] = np.array(adata.obs['cell'])
    adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    sc.pp.normalize_total(adata, target_sum=int(1e6))
    sc.pp.log1p(adata)    
    
    # predict using a pretrained scNym model
    with pytest.raises(NotImplementedError, match='Pretrained'):
        scnym_api(
            adata=adata,
            task='predict',
            groupby='cell',
            out_path=sc.settings.datasetdir,
            trained_model='pretrained_human',
        )
    
#     assert 'X_scnym' in adata.obsm.keys()
#     assert 'scNym' in adata.obs.columns
    
#     # check accuracy
#     # these labels won't match, so we just inspect them manually
#     # to assess sanity
#     target_gt = np.array(
#         adata.obs.loc[target_bidx, 'ground_truth']
#     )
#     pred = np.array(adata.obs.loc[target_bidx, 'scNym'])
#     print('Example Truth vs. Prediction')
#     for i in range(30):
#         print(f'{target_gt[i]}\t|\t{pred[i]}')

    return


def test_atlas2target():
    np.random.seed(1)    
    adata = sc.datasets.paul15()
    sc.pp.normalize_total(adata, target_sum=1e6)
    sc.pp.log1p(adata)
    
    joint_adata = atlas2target(
        adata=adata,
        species='mouse',
        key_added='annotations',
    )
    
    assert 'annotations' in joint_adata.obs.keys()
    assert np.sum(joint_adata.obs['annotations']=='Unlabeled')==adata.shape[0]
    
    return


def test_sslwithatlas():
    np.random.seed(1) 
    adata = sc.datasets.paul15()
    sc.pp.normalize_total(adata, target_sum=1e6)
    sc.pp.log1p(adata)    
    
    joint_adata = atlas2target(
        adata=adata,
        species='mouse',
        key_added='annotations',
    )
    # downsample to speed up the test
    ridx = np.random.choice(
        joint_adata.shape[0],
        size=5000,
        replace=False,
    )
    joint_adata = joint_adata[ridx, :].copy()
    sc.pp.filter_genes(joint_adata, min_cells=10)
    print('Subsampled `joint_adata`: %d cells, %d genes' % joint_adata.shape)
    print('joint obs columns:')
    print(joint_adata.obs.columns)
    print()
    config = {'n_epochs': 2, 'batch_size': 32}
    scnym_api(
        adata=joint_adata,
        task='train',
        groupby='annotations',
        out_path=str(sc.settings.datasetdir),
        config=config,
    )
    return


@cuda_only
def test_user_indices():
    '''Test scNym training and prediction with user supplied indices
    for the data split.'''
    np.random.seed(1)
    # load a testing dataset
    adata = sc.datasets._datasets.read(sc.settings.datasetdir / 'kang17.h5ad', backup_url=TEST_URL)
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['cell'] = np.array(adata.obs['cell'])
    adata.obs['ground_truth'] = np.array(adata.obs['cell'])
    adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    # downsample to speed up testing
    ridx = np.random.choice(
        adata.shape[0],
        size=2048,
        replace=False,
    )
    adata = adata[ridx, :].copy()
    
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['scNym_split'] = np.random.choice(
        ['train', 'test', 'val'],
        size=adata.shape[0],
        p=[0.8, 0.1, 0.1],
        replace=True,
    )
    print('train samples = ', np.sum(adata.obs['scNym_split']=='train'))
    print('test  samples = ', np.sum(adata.obs['scNym_split']=='test'))
    print('val   samples = ', np.sum(adata.obs['scNym_split']=='val'))
    
    # train an scNym model
    config = {'n_epochs': 1}
    scnym_api(
        adata=adata,
        task='train',
        groupby='cell',
        out_path=str(sc.settings.datasetdir),
        config=config,
    )
    
    assert 'scNym_train_results' in adata.uns.keys()
    assert osp.exists(osp.join(str(sc.settings.datasetdir), '00_best_model_weights.pkl'))
    
    # predict cell types with an scNym model
    scnym_api(
        adata=adata,
        task='predict',
        key_added='scNym',
        out_path=str(sc.settings.datasetdir),
        trained_model=str(sc.settings.datasetdir),
        config=config,
    )
    
    assert 'X_scnym' in adata.obsm.keys()
    assert 'scNym' in adata.obs.columns
    
    # check accuracy
    target_gt = np.array(
        adata.obs.loc[target_bidx, 'ground_truth']
    )
    pred = np.array(adata.obs.loc[target_bidx, 'scNym'])
    print('Example Truth vs. Prediction')
    for i in range(10):
        print(f'{target_gt[i]}\t|\t{pred[i]}')
    print()
    acc = np.sum(target_gt == pred) / len(pred) * 100
    print(f'Accuracy on target set: {acc:.03f}%')
    
    return


@cuda_only
def test_user_domains():
    '''Test scNym training and prediction with user supplied domain
    labels for each cell.'''
    np.random.seed(1)
    # load a testing dataset
    adata = sc.datasets._datasets.read(sc.settings.datasetdir / 'kang17.h5ad', backup_url=TEST_URL)
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['cell'] = np.array(adata.obs['cell'])
    adata.obs['ground_truth'] = np.array(adata.obs['cell'])
    adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    # downsample to speed up testing
    ridx = np.random.choice(
        adata.shape[0],
        size=2048,
        replace=False,
    )
    adata = adata[ridx, :].copy()
    
    # use individual patients as unique domain labels
    domain_groupby = 'ind'
    
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['scNym_split'] = np.random.choice(
        ['train', 'test', 'val'],
        size=adata.shape[0],
        p=[0.8, 0.1, 0.1],
        replace=True,
    )
    print('train samples = ', np.sum(adata.obs['scNym_split']=='train'))
    print('test  samples = ', np.sum(adata.obs['scNym_split']=='test'))
    print('val   samples = ', np.sum(adata.obs['scNym_split']=='val'))
    
    # train an scNym model
    config = {'n_epochs': 1, 'domain_groupby': domain_groupby}
    scnym_api(
        adata=adata,
        task='train',
        groupby='cell',
        out_path=str(sc.settings.datasetdir),
        config=config,
    )
    
    assert 'scNym_train_results' in adata.uns.keys()
    assert osp.exists(osp.join(str(sc.settings.datasetdir), '00_best_model_weights.pkl'))
    
    # predict cell types with an scNym model
    scnym_api(
        adata=adata,
        task='predict',
        key_added='scNym',
        out_path=str(sc.settings.datasetdir),
        trained_model=str(sc.settings.datasetdir),
        config=config,
    )
    
    assert 'X_scnym' in adata.obsm.keys()
    assert 'scNym' in adata.obs.columns
    
    # check accuracy
    target_gt = np.array(
        adata.obs.loc[target_bidx, 'ground_truth']
    )
    pred = np.array(adata.obs.loc[target_bidx, 'scNym'])
    print('Example Truth vs. Prediction')
    for i in range(10):
        print(f'{target_gt[i]}\t|\t{pred[i]}')
    print()
    acc = np.sum(target_gt == pred) / len(pred) * 100
    print(f'Accuracy on target set: {acc:.03f}%')
    
    return


@cuda_only
def test_tune():
    '''Test a hyperparameter tuning experiment for scNym'''
    np.random.seed(1)
    # load a testing dataset
    adata = sc.datasets._datasets.read(sc.settings.datasetdir / 'kang17.h5ad', backup_url=TEST_URL)
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['cell'] = np.array(adata.obs['cell'])
    adata.obs['ground_truth'] = np.array(adata.obs['cell'])
    adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    ridx = np.random.choice(adata.shape[0], size=2000, replace=False)
    adata = adata[ridx, :].copy()
    
    parameters = {
        'n_epochs': [1,],
        'lr': [1e-3, 1.],
        'ssl_kwargs': {
            'dan_ramp_epochs': [20,],
        }
    }
    
    results, best_params = scnym_tune(
        adata=adata,
        groupby='cell',
        parameters=parameters,
        hold_out_only=True,
        groupby_eval='ground_truth',
        n_splits=2,
    )
    
    print('Results')
    print('-'*10)
    print(results)
    print('-'*10)
    print()
    
    return


@cuda_only
def test_tune_provide_splits():
    '''Test a hyperparameter tuning experiment for scNym
    with user provided 5-fold splits.
    '''
    np.random.seed(1)
    # load a testing dataset
    adata = sc.datasets._datasets.read(sc.settings.datasetdir / 'kang17.h5ad', backup_url=TEST_URL)
    target_bidx = adata.obs['stim']=='stim'
    adata.obs['cell'] = np.array(adata.obs['cell'])
    adata.obs['ground_truth'] = np.array(adata.obs['cell'])
    adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    ridx = np.random.choice(adata.shape[0], size=2000, replace=False)
    adata = adata[ridx, :].copy()
    
    split_data(
        adata,
        groupby='cell',
        n_splits=2,
    )    
    
    parameters = {
        'n_epochs': [1,],
        'lr': [1e-3, 1.],
        'ssl_kwargs': {
            'dan_ramp_epochs': [20,],
        },
    }
    
    results, best_params = scnym_tune(
        adata=adata,
        groupby='cell',
        parameters=parameters,
        hold_out_only=True,
        groupby_eval='ground_truth',
        n_splits=2,
    )
    
    print('Results')
    print('-'*10)
    print(results)
    print('-'*10)
    print()
    
    return


def main():
    print('<-- test_scnym')
    test_scnym()
    print('<-- test_assumption_checking')
    test_assumption_checking()
    print('<-- test_pretrained')
    test_pretrained()
    print('<-- test_atlas2target')
    test_atlas2target()
    print('<-- test_sslwithatlas')
    test_sslwithatlas()    
    print('<-- test_user_indices')
    test_user_indices()
    print('<-- test_user_domains')
    test_user_domains()
    print('<-- test_tuning')
    test_tune()
    print('<-- test_tune_provide_splits')
    test_tune_provide_splits()
    return


if __name__ == '__main__':
    main()


__author__ = 'Jacob C. Kimmel, David R. Kelley'
__email__ = 'jacobkimmel+scnym@gmail.com, drk@calicolabs.com'
__version__ = '0.3.1'

# populate the namespace so top level imports work
# e.g.
# >> from scnym.model import CellTypeCLF
from . import api, main, dataprep, interpret, model, predict, trainer, utils

